from typing import Tuple

import numpy as np
import scipy.special as spc
import scipy.stats as st

from ..base import Distribution, LinkFunction
from ..links import Logit, Log


class BetaInflatedZero(Distribution):
    """The Zero Inflated Beta Distribution for GAMLSS.
    
    f_Y(y \\mid \\mu, \\sigma, \\nu) = 
    \\begin{cases}
    p_0 & \text{if } y = 0 \\
    (1 - p_0) f_W(y \mid \mu, \sigma) & \\text{if } 0 < y < 1
    \\end{cases}
        
    where  $p_0 = \\nu (1 + \\nu)^{-1}$
    
    and $\\mu, \\sigma \\in (0,1)$ and $\\nu > 0 $
    
    """

    corresponding_gamlss: str = "BEINF0"

    parameter_names = {0: "mu", 1: "sigma", 2: "nu"}
    parameter_support = {
        0: (np.nextafter(0, 1), np.nextafter(1, 0)),
        1: (np.nextafter(0, 1), np.nextafter(1, 0)),
        2: (np.nextafter(0, 1), np.inf),
    }
    distribution_support = (0, np.nextafter(1, 0))

    def __init__(
        self,
        loc_link: LinkFunction = Logit(),
        scale_link: LinkFunction = Logit(),
        inflation_link: LinkFunction = Log(),
    ) -> None:
        super().__init__(links={0: loc_link, 1: scale_link, 2: inflation_link})

    def dl1_dp1(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma, nu = self.theta_to_params(theta)

        if param == 0:
            alpha = mu * (1 - sigma**2) / sigma**2
            beta = (1 - mu) * (1 - sigma**2) / sigma**2

            result = ((1 - sigma**2) / sigma**2) * (
                -spc.digamma(alpha) + spc.digamma(beta) + np.log(y) - np.log(1 - y)
            )

            return np.where(y == 0, 0, result)

        if param == 1:
            alpha = mu * (1 - sigma**2) / sigma**2
            beta = (1 - mu) * (1 - sigma**2) / sigma**2

            result = -(2 / sigma**3) * (
                mu * (-spc.digamma(alpha) + spc.digamma(alpha + beta) + np.log(y))
                + (1 - mu)
                * (-spc.digamma(beta) + spc.digamma(alpha + beta) + np.log(1 - y))
            )

            return np.where(y == 0, 0, result)

        if param == 2:
            return np.where(y == 0, (1 / nu), 0) - (1 / (1 + nu))

    def dl2_dp2(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma, nu = self.theta_to_params(theta)
        if param == 0:
            # MU
            alpha = mu * (1 - sigma**2) / sigma**2
            beta = (1 - mu) * (1 - sigma**2) / sigma**2

            result = -(((1 - sigma**2) ** 2) / sigma**4) * (
                spc.polygamma(1, alpha) + spc.polygamma(1, beta)
            )

            return np.where(y == 0, 0, result)

        if param == 1:
            # SIGMA
            alpha = mu * (1 - sigma**2) / sigma**2
            beta = (1 - mu) * (1 - sigma**2) / sigma**2

            result = -(4 / (sigma**6)) * (
                (mu**2) * spc.polygamma(1, alpha)
                + ((1 - mu) ** 2) * spc.polygamma(1, beta)
                - spc.polygamma(1, alpha + beta)
            )

            return np.where(y == 0, 0, result)

        if param == 2:
            return -1 / (nu * ((1 + nu) ** 2))

    def dl2_dpp(
        self, y: np.ndarray, theta: np.ndarray, params: Tuple[int, int] = (0, 1)
    ) -> np.ndarray:
        self._validate_dl2_dpp_inputs(y, theta, params)
        mu, sigma, nu = self.theta_to_params(theta)

        if sorted(params) == [0, 1]:
            alpha = mu * (1 - sigma**2) / sigma**2
            beta = (1 - mu) * (1 - sigma**2) / sigma**2

            result = (2 * (1 - sigma**2) / sigma**5) * (
                mu * spc.polygamma(1, alpha) - (1 - mu) * spc.polygamma(1, beta)
            )

            return np.where(y == 0, 0, result)

        if sorted(params) == [0, 2]:
            return np.zeros_like(y)  ###

        if sorted(params) == [1, 2]:
            return np.zeros_like(y)  ###

    def initial_values(self, y: np.ndarray) -> np.ndarray:
        return np.tile([np.mean(y), 0.5, np.mean(y)], (y.shape[0], 1))

    def cdf(self, y, theta):
        mu, sigma, nu = self.theta_to_params(theta)
        alpha = mu * (1 - sigma**2) / sigma**2
        beta = alpha * (1 - mu) / mu

        cdf_beta = st.beta(alpha, beta).cdf(y)

        raw_cdf = np.where(y == 0, nu, nu + cdf_beta)

        result = raw_cdf / (1 + nu)

        result = np.where(y < 0, 0, result)
        result = np.where(y >= 1, 1, result)

        return result

    def pdf(self, y, theta):
        mu, sigma, nu = self.theta_to_params(theta)
        alpha = mu * (1 - sigma**2) / sigma**2
        beta = (1 - mu) * (1 - sigma**2) / sigma**2

        pdf_beta = st.beta(alpha, beta, loc=0, scale=1).pdf(y)

        result = np.where(
            (y < 0) | (y >= 1),
            0,
            np.where(y == 0, nu / (1 + nu), pdf_beta / (1 + nu)),
        )

        return result

    def ppf(self, q, theta):
        mu, sigma, nu = self.theta_to_params(theta)

        alpha = mu * (1 - sigma**2) / sigma**2
        beta = (1 - mu) * (1 - sigma**2) / sigma**2

        denom = 1 + nu
        lower = nu / denom

        q = np.asarray(q, dtype=np.float64)
        adjusted_q = (q - lower) * denom  # same as (q - lower) / (1/denom)

        finite_result = st.beta.ppf(adjusted_q, a=alpha, b=beta)

        finite_result = np.where(q <= lower, 0.0, finite_result)
        finite_result = np.where(q == 0, 0.0, finite_result)
        finite_result = np.where(q == 1, np.nextafter(1, 0), finite_result)
        result = np.where((q < 0) | (q > 1), np.nan, finite_result)

        return result

    def pmf(self, y, theta):
        raise NotImplementedError("PMF is not implemented for mixed distributions.")

    def rvs(self, size, theta):
        sim_unif = st.uniform.rvs(size=size)
        return self.ppf(q=sim_unif, theta=theta)

    def logpdf(self, y, theta):
        mu, sigma, nu = self.theta_to_params(theta)
        alpha = mu * (1 - sigma**2) / sigma**2
        beta = (1 - mu) * (1 - sigma**2) / sigma**2

        logpdf_beta = st.beta(alpha, beta).logpdf(y)

        logfy = np.zeros_like(y, dtype=np.float64)
        logfy = np.where(y > 0, logpdf_beta, logfy)
        logfy = np.where(y == 0, np.log(nu), logfy)
        result = logfy - np.log(1 + nu)

        return result

    def logcdf(self, y, theta):
        mu, sigma, nu = self.theta_to_params(theta)
        alpha = mu * (1 - sigma**2) / sigma**2
        beta = (1 - mu) * (1 - sigma**2) / sigma**2

        y = np.asarray(y, dtype=np.float64)
        denom = 1 + nu

        cdf_beta = st.beta(alpha, beta).cdf(y)

        raw_cdf = np.where((y > 0) & (y <= 1), nu + cdf_beta, np.where(y == 0, nu, 0))

        with np.errstate(divide="ignore"):
            log_result = np.log(raw_cdf / denom)

        log_result = np.where(y < 0, -np.inf, log_result)
        log_result = np.where(y >= 1, 0.0, log_result)

        return log_result

    def logpmf(self, y, theta):
        return super().logpmf(y, theta)

    def calculate_conditional_initial_values(self, y, theta, param):
        return super().calculate_conditional_initial_values(y, theta, param)



from typing import Literal

import numpy as np

from .elasticnet import ElasticNetPath


class LassoPath(ElasticNetPath):
    """
    Path-based lasso estimation.

    The lasso method runs coordinate descent along a (geometric) decreasing grid of regularization strengths (lambdas).
    We automatically calculate the maximum regularization strength for which all (not-regularized) coefficients are 0.
    The lower end of the lambda grid is defined as $$\\lambda_\min = \\lambda_\max * \\varepsilon_\\lambda.$$

    We allow to pass user-defined lower and upper bounds for the coefficients.
    The coefficient bounds must be an `numpy` array of the length of `X` respectively of the number of variables in the
    equation _plus the intercept, if you fit one_. This allows to box-constrain the coefficients to a certain range.

    Furthermore, we allow to choose the start value, i.e. whether you want an update to be warm-started on the previous fit's path
    or on the previous reguarlization strength or an average of both. If your data generating process is rather stable,
    the `"previous_fit"` should give considerable speed gains, since warm starting on the previous strength is effectively batch-fitting.

    Lastly, we have some rather technical parameters like the number of coordinate descent iterations,
    whether you want to cycle randomly and for which tolerance you want to break. We use active set iterations, i.e.
    after the first coordinate-wise update for each regularization strength, only non-zero coefficients are updated.

    We use `numba` to speed up the coordinate descent algorithm.
    """

    def __init__(
        self,
        lambda_n: int = 100,
        lambda_eps: float = 1e-4,
        early_stop: int = 0,
        start_value_initial: Literal[
            "previous_lambda", "previous_fit", "average"
        ] = "previous_lambda",
        start_value_update: Literal[
            "previous_lambda", "previous_fit", "average"
        ] = "previous_fit",
        selection: Literal["cyclic", "random"] = "cyclic",
        beta_lower_bound: np.ndarray | None = None,
        beta_upper_bound: np.ndarray | None = None,
        tolerance: float = 1e-4,
        max_iterations: int = 1000,
    ):
        """
        Initializes the lasso method with the specified parameters.

        Args:
            lambda_n (int): Number of lambda values to use in the path. Default is 100.
            lambda_eps (float): Minimum lambda value as a fraction of the maximum lambda. Default is 1e-4.
            early_stop (int): Early stopping criterion. Will stop if the number of non-zero parameters is reached. Default is 0 (no early stopping).
            start_value_initial (Literal["previous_lambda", "previous_fit", "average"]): Method to initialize the start value for the first lambda. Default is "previous_lambda".
            start_value_update (Literal["previous_lambda", "previous_fit", "average"]): Method to update the start value for subsequent lambdas. Default is "previous_fit".
            selection (Literal["cyclic", "random"]): Method to select features during the path. Default is "cyclic".
            beta_lower_bound (np.ndarray | None): Lower bound for the coefficients. Default is None.
            beta_upper_bound (np.ndarray | None): Upper bound for the coefficients. Default is None.
            tolerance (float): Tolerance for the optimization. Default is 1e-4.
            max_iterations (int): Maximum number of iterations for the optimization. Default is 1000.
        """
        super().__init__(
            alpha=1.0,
            lambda_n=lambda_n,
            lambda_eps=lambda_eps,
            early_stop=early_stop,
            start_value_initial=start_value_initial,
            start_value_update=start_value_update,
            selection=selection,
            beta_lower_bound=beta_lower_bound,
            beta_upper_bound=beta_upper_bound,
            tolerance=tolerance,
            max_iterations=max_iterations,
        )



from typing import Tuple

import numpy as np
import scipy.special as spc
import scipy.stats as st

from ..base import Distribution, LinkFunction
from ..links import Logit, Log


class ZeroAdjustedGamma(Distribution):
    """The Zero Adjusted Gamma Distribution for GAMLSS.
    
    The zero adjusted gamma distribution is a mixture of a discrete value 0 with
    probability \\nu, and a gamma GA(\\mu; \\sigma) distribution on the positive real line (0, \\infty)
    with probability (1 - \\nu).
    
    $$
    f_Y(y \\mid \\mu, \\sigma, \\nu) = 
    \\begin{cases}
        \\nu & \\text{if } y = 0 \\
        (1 - \\nu) f_W(y \mid \mu, \sigma) & \\text{if } y > 0
    \end{cases} 
    $$
    
    where $y$ is the observed data, $\\mu > 0$ is the location parameter,
    $\\sigma > 0$ is the scale parameter,
    and $\\nu \\in [0, \\infty) $  is the inflation parameter.

    """

    corresponding_gamlss: str = "ZAGA"
    parameter_names = {0: "mu", 1: "sigma", 2: "nu"}
    parameter_support = {
        0: (np.nextafter(0, 1), np.inf),
        1: (np.nextafter(0, 1), np.inf),
        2: (np.nextafter(0, 1), np.nextafter(1, 0)),
    }
    distribution_support = (0, np.inf)

    def __init__(
        self,
        loc_link: LinkFunction = Log(),
        scale_link: LinkFunction = Log(),
        inflation_link: LinkFunction = Logit(),
    ) -> None:
        super().__init__(links={0: loc_link, 1: scale_link, 2: inflation_link})

    def dl1_dp1(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma, nu = self.theta_to_params(theta)

        if param == 0:
            result = (y - mu) / ((sigma**2) * (mu**2))

            return np.where(y == 0, 0, result)

        if param == 1:
            result = (2 / sigma**3) * (
                (y / mu)
                - np.log(y)
                + np.log(mu)
                + np.log(sigma**2)
                - 1
                + spc.digamma(1 / (sigma**2))
            )

            return np.where(y == 0, 0, result)

        if param == 2:
            return np.where(y == 0, 1 / nu, -1 / (1 - nu))

    def dl2_dp2(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma, nu = self.theta_to_params(theta)
        if param == 0:
            result = -1 / ((sigma**2) * (mu**2))

            return np.where(y == 0, 0, result)

        if param == 1:
            result = (4 / sigma**4) - (4 / sigma**6) * spc.polygamma(1, (1 / sigma**2))

            return np.where(y == 0, 0, result)

        if param == 2:
            return -1 / (nu * (1 - nu))

    def dl2_dpp(
        self, y: np.ndarray, theta: np.ndarray, params: Tuple[int, int] = (0, 1)
    ) -> np.ndarray:
        self._validate_dl2_dpp_inputs(y, theta, params)
        mu, sigma, nu = self.theta_to_params(theta)

        if sorted(params) == [0, 1]:
            return np.zeros_like(y)

        if sorted(params) == [0, 2]:
            return np.zeros_like(y)

        if sorted(params) == [1, 2]:
            return np.zeros_like(y)

    def initial_values(self, y: np.ndarray) -> np.ndarray:
        return np.tile([np.mean(y), 1, 0.5], (y.shape[0], 1))

    def cdf(self, y, theta):
        mu, sigma, nu = self.theta_to_params(theta)
        shape = 1 / sigma**2
        scale = mu * sigma**2
        cdf_gamma = st.gamma(a=shape, loc=0, scale=scale).cdf(y)
        cont_cdf = nu + ((1 - nu) * cdf_gamma)

        result = np.where(y == 0, nu, cont_cdf)

        return result

    def pdf(self, y, theta):
        mu, sigma, nu = self.theta_to_params(theta)
        shape = 1 / sigma**2
        scale = mu * sigma**2
        pdf_gamma = st.gamma(a=shape, loc=0, scale=scale).pdf(y)

        result = (
            (1 - nu)
            * (y / (mu * sigma**2)) ** (1 / sigma**2)
            * (np.exp(-y / (mu * sigma**2)))
            / (y * spc.gamma(1 / sigma**2))
        )

        return np.where(y == 0, nu, result)

    def ppf(self, q, theta):
        mu, sigma, nu = self.theta_to_params(theta)
        shape = 1 / sigma**2
        scale = mu * sigma**2

        result = st.gamma(a=shape, loc=0, scale=scale).ppf((q - nu) / (1 - nu))

        return np.where(q <= nu, 0, result)

    def pmf(self, y, theta):
        raise NotImplementedError("PMF is not implemented for mixed distributions.")

    def rvs(self, size, theta):
        sim_unif = st.uniform.rvs(size=size)
        return self.ppf(q=sim_unif, theta=theta)

    def logpdf(self, y, theta):
        mu, sigma, nu = self.theta_to_params(theta)

        result = (
            np.log(1 - nu)
            + (1 / sigma**2) * np.log(y / (mu * sigma**2))
            - y / (mu * sigma**2)
            - np.log(y)
            - spc.gammaln(1 / sigma**2)
        )

        return np.where(y == 0, np.log(nu), result)

    def logcdf(self, y, theta):
        mu, sigma, nu = self.theta_to_params(theta)
        shape = 1 / sigma**2
        scale = mu * sigma**2
        cdf_gamma = st.gamma(a=shape, loc=0, scale=scale).cdf(y)
        cont_logcdf_gamma = np.log(nu + ((1 - nu) * cdf_gamma))

        result = np.where(y == 0, np.log(nu), cont_logcdf_gamma)

        return result

    def logpmf(self, y, theta):
        return super().logpmf(y, theta)

    def calculate_conditional_initial_values(self, y, theta, param):
        return super().calculate_conditional_initial_values(y, theta, param)



from abc import ABC

import numpy as np
from sklearn.utils.validation import check_is_fitted


class OndilEstimatorMixin(ABC):

    @property
    def is_fitted(self) -> bool:
        """Has the estimator been fitted."""
        return hasattr(self, "n_observations_")

    @property
    def n_samples_(self) -> int:
        check_is_fitted(self, "n_observations_")
        return self.n_observations_

    def partial_fit(
        self, X: np.ndarray, y: np.ndarray, sample_weight: np.ndarray | None = None
    ):
        """
        Align ondil with the scikit-learn API for partial fitting.

        The first partial fit will call `fit`, and subsequent calls will call `update`.
        Allows furthermore to use the sklearn testing framework.

        Parameters
        ----------
        X : np.ndarray
            The input data.
        y : np.ndarray
            The target values.
        sample_weight : np.ndarray, optional
            Sample weights for the observations, by default None.
        Returns
        -------
        self : Estimator
            The fitted estimator.

        """
        if self.is_fitted:
            self.update(X=X, y=y, sample_weight=sample_weight)
        else:
            self.fit(X=X, y=y, sample_weight=sample_weight)
        return self


class Estimator(ABC):

    def __init__(self):
        self.n_observations: int
        self.n_training: float

    @property
    def is_fitted(self) -> bool:
        """Has the estimator been fitted."""
        return self.n_observations > 0

    @property
    def n_obs(self) -> int:
        return self.n_observations

    @property
    def n_train(self) -> float:
        return self.n_training

    @property
    def coef_(self):
        return self.beta

    @property
    def coef_path_(self):
        return self.beta_path

    def fit(
        self,
        X: np.ndarray,
        y: np.ndarray,
        sample_weights: np.ndarray | None = None,
    ) -> None:
        raise NotImplementedError("Not Implemented")

    def update(
        self,
        X: np.ndarray,
        y: np.ndarray,
        sample_weights: np.ndarray | None = None,
    ) -> None:
        raise NotImplementedError("Not Implemented")

    def partial_fit(
        self,
        X: np.ndarray,
        y: np.ndarray,
        sample_weights: np.ndarray | None = None,
    ) -> None:
        self.update(X=X, y=y, sample_weights=sample_weights)



from .beta import Beta
from .betainflated import BetaInflated
from .betainflatedzero import BetaInflatedZero
from .exponential import Exponential
from .gamma import Gamma
from .gumbel import Gumbel
from .inversegamma import InverseGamma
from .inversegaussian import InverseGaussian
from .johnsonsu import JSU
from .logistic import Logistic
from .lognormal import LogNormal
from .lognormalmedian import LogNormalMedian
from .normal import Normal, NormalMeanVariance
from .reversegumbel import ReverseGumbel
from .studentt import StudentT
from .zeroadjustedgamma import ZeroAdjustedGamma

__all__ = [
    "Normal",
    "NormalMeanVariance",
    "StudentT",
    "JSU",
    "BetaInflated",
    "Gamma",
    "Beta",
    "LogNormal",
    "LogNormalMedian",
    "Logistic",
    "Exponential",
    "Gumbel",
    "InverseGaussian",
    "ReverseGumbel",
    "InverseGamma",
    "BetaInflatedZero",
    "ZeroAdjustedGamma",
]



# ruff: noqa: E402

from importlib.metadata import version
from importlib.util import find_spec

HAS_PANDAS = False
HAS_POLARS = False

if find_spec("pandas") is not None:
    HAS_PANDAS = True

if find_spec("polars") is not None:
    HAS_POLARS = True

from .information_criteria import InformationCriterion
from .scaler import OnlineScaler

__version__ = version("ondil")

__all__ = [
    "OnlineScaler",
    "InformationCriterion",
]



from .identitylinks import Identity
from .logitlinks import Logit
from .loglinks import LogIdent, Log, LogShiftTwo, LogShiftValue
from .softpluslinks import (
    InverseSoftPlus,
    InverseSoftPlusShiftTwo,
    InverseSoftPlusShiftValue,
)
from .sqrtlinks import Sqrt, SqrtShiftTwo, SqrtShiftValue


__all__ = [
    "Identity",
    "InverseSoftPlus",
    "InverseSoftPlusShiftTwo",
    "InverseSoftPlusShiftValue",
    "LogIdent",
    "Logit",
    "Log",
    "LogShiftTwo",
    "LogShiftValue",
    "Sqrt",
    "SqrtShiftTwo",
    "SqrtShiftValue",
]



from .online_gamlss import OnlineDistributionalRegression
from .online_lasso import OnlineLasso
from .online_linear_model import OnlineLinearModel

__all__ = ["OnlineDistributionalRegression", "OnlineLasso", "OnlineLinearModel"]



import numbers

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin, _fit_context
from sklearn.utils._param_validation import Interval
from sklearn.utils.validation import (
    _check_sample_weight,
    check_is_fitted,
    validate_data,
)

from .base.estimator import OndilEstimatorMixin
from .gram import init_forget_vector


class OnlineScaler(OndilEstimatorMixin, TransformerMixin, BaseEstimator):

    _parameter_constraints = {
        "forget": [Interval(numbers.Real, 0.0, 1.0, closed="left")],
        "to_scale": [bool, np.ndarray],
    }

    def __init__(
        self,
        forget: float = 0.0,
        to_scale: bool | np.ndarray = True,
    ):
        """The online scaler allows for incremental updating and scaling of matrices.

        Args:
            forget (float, optional): The forget factor. Older observations will be exponentially discounted. Defaults to 0.0.
            to_scale (bool | np.ndarray, optional): The variables to scale.
                `True` implies all variables will be scaled.
                `False` implies no variables will be scaled.
                An `np.ndarray` of type `bool` or `int` implies that the columns `X[:, to_scale]` will be scaled, all other columns will not be scaled.
                Defaults to True.
        """
        self.forget = forget
        self.to_scale = to_scale

    def _prepare_estimator(self, X: np.ndarray):
        """Add derived attributes to estimator"""
        if isinstance(self.to_scale, np.ndarray):
            self._selection = self.to_scale
            self._do_scale = True
        elif isinstance(self.to_scale, bool):
            if self.to_scale:
                self._selection = np.arange(X.shape[1])
                self._do_scale = True
            else:
                self._selection = False
                self._do_scale = False

        # Variables
        self.mean_ = 0
        self.var_ = 0
        self._M = 0
        self._cumulative_w = 0  # Track cumulative weights for exponential forgetting

    @property
    def std_(self) -> float | np.ndarray:
        """Standard deviation of the scaled variables."""
        check_is_fitted(self, ["mean_", "var_"])
        if self._do_scale:
            return np.sqrt(self.var_)
        else:
            return 1.0

    @_fit_context(prefer_skip_nested_validation=True)
    def fit(
        self,
        X: np.ndarray,
        y: None = None,
        sample_weight: np.ndarray | None = None,
    ) -> "OnlineScaler":
        """Fit the OnlineScaler() Object for the first time.

        Args:
            X (np.ndarray): Matrix of covariates X.
            y (None, optional): Not used, present for compatibility with sklearn API. Defaults to None.
            sample_weight (np.ndarray, optional): Weights for each sample. Defaults to None (uniform weights).
        """

        X = validate_data(
            self,
            X=X,
            y=None,
            reset=True,
            ensure_min_samples=2,
            dtype=[np.float64, np.float32],
        )
        sample_weight = _check_sample_weight(X=X, sample_weight=sample_weight)
        self._prepare_estimator(X)
        self.n_observations_ = sample_weight.sum()

        if self._do_scale:
            forget_vector = init_forget_vector(self.forget, X.shape[0])
            effective_weights = sample_weight * forget_vector

            self._cumulative_w = np.sum(
                effective_weights
            )  # Initialize cumulative weight
            self.mean_ = np.average(
                X[:, self._selection], weights=effective_weights, axis=0
            )

            # Calculate the variance of each column of x_init and assing it to self.var_
            diff_sq = (X[:, self._selection] - self.mean_) ** 2
            self.var_ = np.average(diff_sq, weights=effective_weights, axis=0)
            self._M = self.var_ * self._cumulative_w

        return self

    @_fit_context(prefer_skip_nested_validation=True)
    def update(self, X: np.ndarray, y=None, sample_weight: np.ndarray = None):
        """Update the `OnlineScaler()` for new rows of X.

        Args:
            X (np.ndarray): New data for X.
            y (None, optional): Not used, present for compatibility with sklearn API. Defaults to None.
            sample_weight (np.ndarray, optional): Weights for each sample. Defaults to None (uniform weights).
        """
        check_is_fitted(self, ["mean_", "var_"])
        X = validate_data(
            self,
            X=X,
            y=None,
            reset=False,
            ensure_min_samples=1,
            dtype=[np.float64, np.float32],
        )
        sample_weight = _check_sample_weight(X=X, sample_weight=sample_weight)
        self.n_observations_ += sample_weight.sum()

        # Loop over all rows of new X
        if self._do_scale:
            for i in range(X.shape[0]):
                # Effective weight for the old state
                eff_old_w = self._cumulative_w * (1 - self.forget)
                self._cumulative_w = eff_old_w + sample_weight[i]
                diff_old = X[i, self._selection] - self.mean_

                # Update mean
                self.mean_ = (
                    self.mean_ * eff_old_w + X[i, self._selection] * sample_weight[i]
                ) / self._cumulative_w

                diff_new = X[i, self._selection] - self.mean_

                # Update M (sum of squared deviations)
                self._M = (
                    self._M * (1 - self.forget) + sample_weight[i] * diff_old * diff_new
                )

                # Update variance
                self.var_ = self._M / self._cumulative_w

        else:
            pass

        return self

    def transform(self, X: np.ndarray) -> np.ndarray:
        """Transform X to a mean-std scaled matrix.

        Args:
            X (np.ndarray): X matrix for covariates.

        Returns:
            np.ndarray: Scaled X matrix.
        """
        check_is_fitted(self, ["mean_", "var_"])
        X = validate_data(
            self,
            X=X,
            y=None,
            reset=False,
            ensure_min_samples=1,
            dtype=[np.float64, np.float32],
        )

        if self._do_scale:
            out = np.copy(X)
            out[:, self._selection] = (X[:, self._selection] - self.mean_) / np.sqrt(
                self.var_
            )
            return out
        else:
            return X

    def inverse_transform(self, X: np.ndarray) -> np.ndarray:
        """Back-transform a scaled X matrix to the original domain.

        Args:
            X (np.ndarray): Scaled X matrix.

        Returns:
            np.ndarray: Scaled back to the original scale.
        """
        check_is_fitted(self, ["mean_", "var_"])
        X = validate_data(
            self,
            X=X,
            reset=False,
            ensure_min_samples=1,
            dtype=[np.float64, np.float32],
        )

        if self._do_scale:
            out = np.copy(X)
            out[:, self._selection] = (
                X[:, self._selection] * np.sqrt(self.var_) + self.mean_
            )
            return out
        else:
            return X



import numba as nb
import numpy as np


@nb.njit()
def init_forget_vector(forget: float, size: int) -> np.ndarray:
    """Initialise an exponentially discounted vector of weights.

    Recursively initialise a vector of exponentially discounted weights of `size` N.

    The weight for $n$-th observation is defined as $(1 - \\text{forget})^{(N - n)}$

    Note that this functions assumes that the first observation is the oldest observation
    and the last observation is the newest observation. This is in line with the standard
    `pandas` way of sorting `pd.DataFrame`s with `Datetime`-indices.

    !!! numba "Numba"
        This function uses `numba` just-in-time-compilation.

    Args:
        forget (float): Forget factor.
        size (int): Length of the output vector.

    Returns:
        np.ndarray: Vector of exponentially discounted weights.
    """
    gamma = 1 - forget
    if gamma == 1:
        out = np.ones(size)
    else:
        out = np.empty(size)
        out[-1] = 1
        for i in range(size - 1, 0, -1):
            out[i - 1] = out[i] * gamma
    return out


@nb.njit()
def init_gram(X: np.ndarray, w: np.ndarray, forget: float = 0) -> np.ndarray:
    """Initialise the Gramian Matrix.

    The Gramian Matrix is defined as
    $$
    G = X^T \Gamma WX
    $$
    where $X$ is the design matrix, $W$ is a diagonal, user-defined weight matrix, $\\Gamma$ is a
    diagonal matrix of exponentially discounting weights.

    !!! numba "Numba"
        This function uses `numba` just-in-time-compilation.

    Args:
        X (np.ndarray): Design matrix $X$
        w (np.ndarray): Weights vector
        forget (float, optional): Forget factor. Defaults to 0.

    Returns:
        np.ndarray: Gramian Matrix.
    """
    f = init_forget_vector(forget, X.shape[0])
    gram = (X * np.expand_dims((w * f) ** 0.5, -1)).T @ (
        X * np.expand_dims((w * f) ** 0.5, -1)
    )
    return gram


@nb.njit()
def init_y_gram(
    X: np.ndarray, y: np.ndarray, w: np.ndarray, forget: float = 0
) -> np.ndarray:
    """Initialise the y-Gramian Matrix.

    The Gramian Matrix is defined as $$
    H = X^T \\Gamma WY
    $$ where $X$ is the design matrix, $Y$ is the response variable, $W$ is a diagonal,
    user-defined weight matrix, $\\Gamma$ is a diagonal matrix of exponentially discounting weights.

    !!! numba "Numba"
        This function uses `numba` just-in-time-compilation.

    Args:
        X (np.ndarray): Design matrix $X$
        y (np.ndarray): Response variable $Y$
        w (np.ndarray): Weights vector
        forget (float, optional): Forget factor. Defaults to 0.

    Returns:
        np.ndarray: y-Gramian Matrix.
    """
    f = init_forget_vector(forget, X.shape[0])
    gram = np.expand_dims(
        (X * np.expand_dims((w * f) ** 0.5, -1)).T @ (y * (w * f) ** 0.5), axis=-1
    )
    return gram


@nb.njit()
def init_inverted_gram(X: np.ndarray, w: np.ndarray, forget: float = 0) -> np.ndarray:
    """Initialise the inverted Gramian Matrix.

    The inverted Gramian Matrix is defined as $$
    G = (X^T \Gamma WX)^{-1}
    $$ where $X$ is the design matrix, $W$ is a diagonal, user-defined weight matrix,
    $\Gamma$ is a diagonal matrix of exponentially discounting weights.

    !!! numba "Numba"
        This function uses `numba` just-in-time-compilation.

    Args:
        X (np.ndarray): Design matrix $X$
        w (np.ndarray): Weights vector
        forget (float, optional): Forget factor. Defaults to 0.

    Returns:
        np.ndarray: Gramian Matrix.

    Raises:
        ValueError: If the matrix is not invertible (if rank(X.T @ X) < X.shape[0]).

    """
    gram = init_gram(X=X, w=w, forget=forget)
    rank = np.linalg.matrix_rank(gram)
    if rank == gram.shape[0]:
        inv_gram = np.linalg.inv(gram)
        return inv_gram
    else:
        raise ValueError("Matrix is not invertible.")


# TODO (SH): For some reason we switched the syntax in C++ here
# We should change that


@nb.njit()
def update_gram(
    gram: np.ndarray, X: np.ndarray, forget: float = 0, w: float = 1
) -> np.ndarray:
    """Update the Gramian Matrix.

    !!! numba "Numba"
        This function uses `numba` just-in-time-compilation.

    Args:
        gram (np.ndarray): Gramian Matrix
        X (np.ndarray): New observations for $X$
        forget (float, optional): Forget factor. Defaults to 0.
        w (float, optional): Weights for the new observations. Defaults to 1.

    Returns:
        np.ndarray: Updated Gramian Matrix.
    """
    if X.shape[0] == 1:
        # Single Step Update
        new_gram = (1 - forget) * gram + w * np.outer(X, X)
    else:
        # Batch Update
        batch_size = X.shape[0]
        f = init_forget_vector(size=batch_size, forget=forget)
        weights = np.expand_dims((w * f) ** 0.5, axis=-1)
        new_gram = gram * (1 - forget) ** batch_size + (X * weights).T @ (X * weights)
    return new_gram


@nb.njit()
def update_y_gram(
    gram: np.ndarray, X: np.ndarray, y: np.ndarray, forget: float = 0, w: float = 1
) -> np.ndarray:
    """Update the Y-Gramian Matrix.

    !!! numba "Numba"
        This function uses `numba` just-in-time-compilation.

    Args:
        gram (np.ndarray): Gramian Matrix
        X (np.ndarray): New Observations for $X$
        y (np.ndarray): New Observations for $Y$
        forget (float, optional): Forget Factor. Defaults to 0.
        w (float, optional): Weights for the new observations. Defaults to 1.

    Returns:
        np.ndarray: Updated Gramian Matrix.
    """
    if X.shape[0] == 1:
        # Single Update
        new_gram = (1 - forget) * gram + w * np.outer(X, y)
    else:
        # Batch update
        batch_size = X.shape[0]
        f = init_forget_vector(size=batch_size, forget=forget)
        new_gram = gram * (1 - forget) ** batch_size + np.expand_dims(
            ((X * np.expand_dims((w * f) ** 0.5, axis=-1)).T @ (y * (w * f) ** 0.5)), -1
        )
    return new_gram


@nb.jit()
def _update_inverted_gram(
    gram: np.ndarray, X: np.ndarray, forget: float = 0, w: float = 1
) -> np.ndarray:
    """Update the inverted Gramian for one step"""
    gamma = 1 - forget
    new_gram = (1 / gamma) * (
        gram - ((w * gram @ np.outer(X, X) @ gram) / (gamma + w * X @ gram @ X.T))
    )
    return new_gram


def update_inverted_gram(
    gram: np.ndarray, X: np.ndarray, forget: float = 0, w: float = 1
) -> np.ndarray:
    """Update the inverted Gramian Matrix.

    !!! numba "Numba"
        This function uses `numba` just-in-time-compilation.

    Args:
        gram (np.ndarray): Inverted Gramian Matrix.
        X (np.ndarray): New observations for $X$.
        forget (float, optional): Forget Factor. Defaults to 0.
        w (float, optional): Weights for new observations. Defaults to 1.

    Returns:
        np.ndarray: Updated inverted Gramian matrix.
    """
    if X.shape[0] == 1:
        new_gram = _update_inverted_gram(gram, X, forget=forget, w=w)
    else:
        new_gram = _update_inverted_gram(
            gram, X=np.expand_dims(X[0, :], 0), forget=forget, w=w[0]
        )
        for i in range(1, X.shape[0]):
            new_gram = _update_inverted_gram(
                gram=new_gram, X=np.expand_dims(X[i, :], 0), forget=forget, w=w[i]
            )
    return new_gram



from typing import Tuple

import numpy as np

from ..base import LinkFunction
from .robust_math import SMALL_NUMBER, robust_exp, robust_log


class Log(LinkFunction):
    """
    The log-link function.

    The log-link function is defined as \(g(x) = \log(x)\).
    """

    link_support = (np.nextafter(0, 1), np.inf)

    def __init__(self):
        pass

    def link(self, x: np.ndarray) -> np.ndarray:
        return robust_log(x)

    def inverse(self, x: np.ndarray) -> np.ndarray:
        return robust_exp(x)

    def inverse_derivative(self, x: np.ndarray) -> np.ndarray:
        return robust_exp(x)

    def link_derivative(self, x: np.ndarray) -> np.ndarray:
        return 1 / x

    def link_second_derivative(self, x: np.ndarray) -> np.ndarray:
        return -1 / x**2


class LogShiftValue(LinkFunction):
    """
    The Log-Link function shifted to a value \(v\).

    This link function is defined as \(g(x) = \log(x - v)\). It can be used
    to ensure that certain distribution paramters don't fall below lower
    bounds, e.g. ensuring that the degrees of freedom of a Student's t distribtuion
    don't fall below 2, hence ensuring that the variance exists.
    """

    def __init__(self, value: float):
        self.value = value

    @property
    def link_support(self) -> Tuple[float, float]:
        return (self.value + np.nextafter(0, 1), np.inf)

    def link(self, x: np.ndarray) -> np.ndarray:
        return robust_log(x - self.value + SMALL_NUMBER)

    def inverse(self, x: np.ndarray) -> np.ndarray:
        return self.value + robust_exp(x)

    def inverse_derivative(self, x: np.ndarray) -> np.ndarray:
        return robust_exp(x)

    def link_derivative(self, x: np.ndarray) -> np.ndarray:
        return 1 / (x - self.value + SMALL_NUMBER)

    def link_second_derivative(self, x: np.ndarray) -> np.ndarray:
        return -1 / (x - self.value + SMALL_NUMBER) ** 2


class LogShiftTwo(LogShiftValue):
    """
    The Log-Link function shifted to 2.

    This link function is defined as \(g(x) = \log(x - 2)\). It can be used
    to ensure that certain distribution paramters don't fall below lower
    bounds, e.g. ensuring that the degrees of freedom of a Student's t distribtuion
    don't fall below 2, hence ensuring that the variance exists.
    """

    def __init__(self):
        super().__init__(2)
        pass


class LogIdent(LinkFunction):
    """The Logident Link function.

    The LogIdent Link function has been introduced by [Narajewski & Ziel 2020](https://arxiv.org/pdf/2005.01365) and can be
    used to avoid the exponential inverse for large values while keeping the log-behaviour in small ranges. This can stabilize
    the estimation procedure.
    """

    link_support = (np.nextafter(0, 1), np.inf)

    def __init__(self):
        pass

    def link(self, x: np.ndarray) -> np.ndarray:
        return np.where(x <= 1, robust_log(x), x - 1)

    def inverse(self, x: np.ndarray) -> np.ndarray:
        return np.where(x <= 0, robust_exp(x), x + 1)

    def inverse_derivative(self, x: np.ndarray) -> np.ndarray:
        return np.where(x <= 0, robust_exp(x), 1)

    def link_derivative(self, x: np.ndarray) -> np.ndarray:
        return super().link_derivative(x)

    def link_second_derivative(self, x) -> np.ndarray:
        return super().link_second_derivative(x)



import numpy as np
import scipy.stats as st

from ..base import Distribution, LinkFunction, ScipyMixin
from ..links import Log


class InverseGaussian(ScipyMixin, Distribution):
    """
    Inverse Gaussian (Wald) distribution for GAMLSS.

    This distribution is characterized by two parameters:
    - $\\mu$: the mean of the distribution.
    - $\\sigma$: the scale parameter, which is related to the variance.

    The probability density function (PDF) is given by:
    $$
        f(y; \\mu, \\sigma) = \\sqrt{\\frac{\\sigma}{2\\pi y^3}} \\exp\\left(-\\frac{(y - \\mu)^2}{2\\sigma^2 y}\\right)
    $$
    where $y > 0$, $\\mu > 0$, and $\\sigma > 0$.

    Note that the Inverse Gaussian distribution in `scipy.stats` is parameterized differently:

    * `mu` is the mean of the distribution.
    * `scale` is the scale parameter

    and the PDF is given by:
    $$
        f(y; \\mu, \\lambda) = \\sqrt{\\frac{\\lambda}{2\\pi y^3}} \\exp\\left(-\\frac{\\lambda (y - \\mu)^2}{2\\mu^2 y}\\right)
    $$
    where $y > 0$, $\\mu > 0$, and $\\lambda > 0$.

    The relationship between the parameters is:

    * `mu` in `scipy.stats` corresponds to $\\mu \\sigma^2$ in this implementation,
    * `scale` in `scipy.stats` corresponds to $1 / \\sigma^2$ in this implementation.
    * The `loc` parameter in `scipy.stats` is always 0.

    """

    corresponding_gamlss: str = "IG"
    parameter_names = {0: "mu", 1: "sigma"}
    parameter_support = {
        0: (np.nextafter(0, 1), np.inf),
        1: (np.nextafter(0, 1), np.inf),
    }
    distribution_support = (np.nextafter(0, 1), np.inf)
    scipy_dist = st.invgauss
    scipy_names = {"mu": "mu", "sigma": "scale"}

    def __init__(
        self,
        loc_link: LinkFunction = Log(),
        scale_link: LinkFunction = Log(),
    ) -> None:
        super().__init__(
            links={
                0: loc_link,
                1: scale_link,
            }
        )

    def theta_to_scipy_params(self, theta: np.ndarray) -> dict:
        mu, sigma = self.theta_to_params(theta)
        return {"mu": mu * sigma**2, "loc": np.zeros_like(mu), "scale": 1 / sigma**2}

    def dl1_dp1(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        mu, sigma = self.theta_to_params(theta)
        if param == 0:
            # (y-mu)/((sigma^2)*(mu^3))
            return (y - mu) / (sigma**2 * mu**3)
        if param == 1:
            # (-1/sigma) +((y-mu)^2)/(y*(sigma^3)*(mu^2)),
            return (-1 / sigma) + ((y - mu) ** 2) / (sigma**3 * mu**2 * y)
        raise ValueError("param must be 0 (mu) or 1 (sigma)")

    def dl2_dp2(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        mu, sigma = self.theta_to_params(theta)
        if param == 0:
            # -1/((mu^3)*(sigma^2)),
            return -1 / (mu**3 * sigma**2)
        if param == 1:
            # -2/(sigma^2),
            return -2 / (sigma**2)
        raise ValueError("param must be 0 (mu) or 1 (sigma)")

    def dl2_dpp(
        self, y: np.ndarray, theta: np.ndarray, params: tuple[int, int] = (0, 1)
    ) -> np.ndarray:
        if params[0] == params[1]:
            raise ValueError("Cross derivatives must use different parameters.")
        return np.zeros_like(y)

    def initial_values(self, y: np.ndarray) -> np.ndarray:
        mu_init = np.mean(y)
        sigma_init = np.std(y) / np.sqrt(np.mean(y))
        initial_params = [mu_init, sigma_init]
        return np.tile(initial_params, (y.shape[0], 1))



import numpy as np
from ondil.coordinate_descent import online_coordinate_descent_path
from sklearn.datasets import load_diabetes
from sklearn.linear_model import lasso_path


def test_coordinate_descent():
    # Sklearn uses a different stopping criterion than we do
    # Also slightly different definition of alpha compared to lambda
    # Hence you get some inconsistencies around _some_ betas in the path

    # Therefore we check the path and assert that 95% of betas match
    # for each explanatory variable

    X, y = load_diabetes(return_X_y=True)
    X /= X.std(axis=0)
    J = X.shape[1]

    x_gram = X.T @ X
    y_gram = X.T @ y
    tolerance = 1e-8

    lambda_eps = 0.0001
    lambda_n = 200
    lambda_max = np.abs(y_gram).max()
    lambda_grid = np.geomspace(lambda_max, lambda_eps * lambda_max, lambda_n)

    is_regularized = np.repeat(True, J)
    beta_path = np.zeros((lambda_n, J))

    ondil_lasso_path = online_coordinate_descent_path(
        x_gram=x_gram,
        y_gram=y_gram,
        beta_path=beta_path,
        lambda_path=lambda_grid,
        alpha=1.0,
        is_regularized=is_regularized,
        early_stop=0,
        beta_lower_bound=np.repeat(-np.inf, J),
        beta_upper_bound=np.repeat(np.inf, J),
        which_start_value="previous_lambda",
        selection="cyclic",
        tolerance=tolerance,
        max_iterations=1000,
    )[0]

    alphas, sklearn_lasso_path, _ = lasso_path(
        X, y, eps=lambda_eps, n_alphas=lambda_n, precompute=x_gram, tol=tolerance
    )
    sklearn_lasso_path = sklearn_lasso_path.T

    assert np.allclose(alphas, lambda_grid / X.shape[0]), "Lambdas don't match"
    assert np.all(
        np.mean(np.isclose(sklearn_lasso_path - ondil_lasso_path, 0), axis=0) > 0.95
    ), "Betas don't match"


def test_coordinate_descent_bounds():
    X, y = load_diabetes(return_X_y=True)
    X /= X.std(axis=0)
    J = X.shape[1]

    x_gram = X.T @ X
    y_gram = X.T @ y
    tolerance = 1e-8

    lambda_eps = 0.0001
    lambda_n = 200
    lambda_max = np.abs(y_gram).max()
    lambda_grid = np.geomspace(lambda_max, lambda_eps * lambda_max, lambda_n)

    is_regularized = np.repeat(True, J)
    beta_path = np.zeros((lambda_n, J))

    ondil_lasso_path_positive = online_coordinate_descent_path(
        x_gram=x_gram,
        y_gram=y_gram,
        beta_path=beta_path,
        lambda_path=lambda_grid,
        alpha=1.0,
        early_stop=0,
        is_regularized=is_regularized,
        beta_lower_bound=np.repeat(0, J),
        beta_upper_bound=np.repeat(np.inf, J),
        which_start_value="previous_lambda",
        selection="cyclic",
        tolerance=tolerance,
        max_iterations=1000,
    )[0]

    ondil_lasso_path_negative = online_coordinate_descent_path(
        x_gram=x_gram,
        y_gram=y_gram,
        beta_path=beta_path,
        lambda_path=lambda_grid,
        alpha=1.0,
        is_regularized=is_regularized,
        early_stop=0,
        beta_lower_bound=np.repeat(-np.inf, J),
        beta_upper_bound=np.repeat(0, J),
        which_start_value="previous_lambda",
        selection="cyclic",
        tolerance=tolerance,
        max_iterations=1000,
    )[0]

    assert np.all(ondil_lasso_path_negative <= 0), "Path should contain only betas <= 0"
    assert np.all(ondil_lasso_path_positive >= 0), "Path should contain only betas >= 0"



from typing import Literal

import numpy as np

from ..base import EstimationMethod
from ..coordinate_descent import online_coordinate_descent_path
from ..gram import init_gram, init_y_gram, update_gram, update_y_gram


class ElasticNetPath(EstimationMethod):
    """
    Path-based elastic net estimation.

    The elastic net method runs coordinate descent along a (geometric) decreasing grid of regularization strengths (lambdas).
    We automatically calculate the maximum regularization strength for which all (not-regularized) coefficients are 0.
    The lower end of the lambda grid is defined as $$\\lambda_\min = \\lambda_\max * \\varepsilon_\\lambda.$$

    The elastic net method is a combination of LASSO and Ridge regression. Parameter $\alpha$ controls the balance between LASSO and Ridge. Thereby, $\alpha=0$ corresponds to Ridge regression and $\alpha=1$ corresponds to LASSO regression.

    We allow to pass user-defined lower and upper bounds for the coefficients.
    The coefficient bounds must be an `numpy` array of the length of `X` respectively of the number of variables in the
    equation _plus the intercept, if you fit one_. This allows to box-constrain the coefficients to a certain range.

    Furthermore, we allow to choose the start value, i.e. whether you want an update to be warm-started on the previous fit's path
    or on the previous reguarlization strength or an average of both. If your data generating process is rather stable,
    the `"previous_fit"` should give considerable speed gains, since warm starting on the previous strength is effectively batch-fitting.

    Lastly, we have some rather technical parameters like the number of coordinate descent iterations,
    whether you want to cycle randomly and for which tolerance you want to break. We use active set iterations, i.e.
    after the first coordinate-wise update for each regularization strength, only non-zero coefficients are updated.

    We use `numba` to speed up the coordinate descent algorithm.

    """

    def __init__(
        self,
        alpha: float,
        lambda_n: int = 100,
        lambda_eps: float = 1e-4,
        early_stop: int = 0,
        start_value_initial: Literal[
            "previous_lambda", "previous_fit", "average"
        ] = "previous_lambda",
        start_value_update: Literal[
            "previous_lambda", "previous_fit", "average"
        ] = "previous_fit",
        selection: Literal["cyclic", "random"] = "cyclic",
        beta_lower_bound: np.ndarray | None = None,
        beta_upper_bound: np.ndarray | None = None,
        tolerance: float = 1e-4,
        max_iterations: int = 1000,
    ):
        """
        Initializes the ElasticNet method with the specified parameters.

        Args:
            alpha (float): Mixing parameter between the L1 and L2 loss. Alpha = 0 corresponds to the Rigde, Alpha = 1 corresponds to the LASSO.
            lambda_n (int): Number of lambda values to use in the path. Default is 100.
            lambda_eps (float): Minimum lambda value as a fraction of the maximum lambda. Default is 1e-4.
            early_stop (int): Early stopping criterion. Will stop if the number of non-zero parameters is reached. Default is 0 (no early stopping).
            start_value_initial (Literal["previous_lambda", "previous_fit", "average"]): Method to initialize the start value for the first lambda. Default is "previous_lambda".
            start_value_update (Literal["previous_lambda", "previous_fit", "average"]): Method to update the start value for subsequent lambdas. Default is "previous_fit".
            selection (Literal["cyclic", "random"]): Method to select features during the path. Default is "cyclic".
            beta_lower_bound (np.ndarray | None): Lower bound for the coefficients. Default is None.
            beta_upper_bound (np.ndarray | None): Upper bound for the coefficients. Default is None.
            tolerance (float): Tolerance for the optimization. Default is 1e-4.
            max_iterations (int): Maximum number of iterations for the optimization. Default is 1000.
        """
        super().__init__(
            _path_based_method=True,
            _accepts_bounds=True,
            _accepts_selection=True,
        )
        self.early_stop = early_stop
        self.beta_lower_bound = beta_lower_bound
        self.beta_upper_bound = beta_upper_bound
        self.lambda_n = lambda_n
        self.lambda_eps = lambda_eps
        self.alpha = alpha
        self.start_value_initial = start_value_initial
        self.start_value_update = start_value_update
        self.selection = selection
        self.tolerance = tolerance
        self.max_iterations = max_iterations

        self._path_length = self.lambda_n

    @staticmethod
    def _get_lambda_max(x_gram, y_gram, is_regularized):
        if np.all(is_regularized):
            lambda_max = np.max(y_gram)
        elif np.sum(~is_regularized) == 1:
            intercept = y_gram[~is_regularized] / np.diag(x_gram)[~is_regularized]
            lambda_max = np.max(
                np.abs(y_gram.flatten() - x_gram[~is_regularized, :] * intercept)
            )
        else:
            raise NotImplementedError(
                "More than one not regularized value is currently not supported."
            )
        return lambda_max

    def _set_and_validate_bounds(self, x_gram: np.ndarray) -> None:
        J = x_gram.shape[1]
        if self.beta_lower_bound is None:
            self.beta_lower_bound = np.repeat(-np.inf, J)
        if self.beta_upper_bound is None:
            self.beta_upper_bound = np.repeat(np.inf, J)

        if len(self.beta_lower_bound) != J:
            raise ValueError("Lower bound does not have correct length")
        if len(self.beta_upper_bound) != J:
            raise ValueError("Upper bound does not have correct length")

    @staticmethod
    def init_x_gram(X, weights, forget):
        return init_gram(X=X, w=weights, forget=forget)

    @staticmethod
    def init_y_gram(X, y, weights, forget):
        return init_y_gram(X, y, w=weights, forget=forget)

    @staticmethod
    def update_x_gram(gram, X, weights, forget):
        return update_gram(gram, X, w=weights, forget=forget)

    @staticmethod
    def update_y_gram(gram, X, y, weights, forget):
        return update_y_gram(gram, X, y, forget=forget, w=weights)

    def fit_beta_path(self, x_gram, y_gram, is_regularized):
        self._set_and_validate_bounds(x_gram=x_gram)
        lambda_max = self._get_lambda_max(
            x_gram=x_gram, y_gram=y_gram, is_regularized=is_regularized
        )
        lambda_path = np.geomspace(
            lambda_max, lambda_max * self.lambda_eps, self.lambda_n
        )
        beta_path = np.zeros((self.lambda_n, x_gram.shape[0]))
        beta_path, _ = online_coordinate_descent_path(
            x_gram=x_gram,
            y_gram=y_gram.squeeze(-1),
            early_stop=self.early_stop,
            beta_path=beta_path,
            lambda_path=lambda_path,
            alpha=self.alpha,
            is_regularized=is_regularized,
            beta_lower_bound=self.beta_lower_bound,
            beta_upper_bound=self.beta_upper_bound,
            which_start_value=self.start_value_initial,
            selection=self.selection,
            tolerance=self.tolerance,
            max_iterations=self.max_iterations,
        )
        return beta_path

    def update_beta_path(self, x_gram, y_gram, beta_path, is_regularized):
        lambda_max = self._get_lambda_max(
            x_gram=x_gram, y_gram=y_gram, is_regularized=is_regularized
        )
        lambda_path = np.geomspace(
            lambda_max, lambda_max * self.lambda_eps, self.lambda_n
        )
        beta_path, _ = online_coordinate_descent_path(
            x_gram=x_gram,
            y_gram=y_gram.squeeze(-1),
            beta_path=beta_path,
            lambda_path=lambda_path,
            alpha=self.alpha,
            early_stop=self.early_stop,
            is_regularized=is_regularized,
            beta_lower_bound=self.beta_lower_bound,
            beta_upper_bound=self.beta_upper_bound,
            which_start_value=self.start_value_update,
            selection=self.selection,
            tolerance=self.tolerance,
            max_iterations=self.max_iterations,
        )
        return beta_path

    def fit_beta(self, x_gram, y_gram, is_regularized):
        pass

    def update_beta(self, x_gram, y_gram, is_regularized):
        pass



from typing import Optional, Tuple

import numpy as np
import scipy.stats as st

from ..base import Distribution, LinkFunction, ScipyMixin
from ..links import Log


class LogNormalMedian(ScipyMixin, Distribution):
    """
    The Log-Normal distribution with median and standard deviation parameterization in the log-space.

    The probability density function of the distribution is defined as:
    $$
    f(y | \\mu, \\sigma) = \\frac{1}{y\\sigma\\sqrt{2\\pi}} \\exp\\left(-\\frac{(\\log y - \\log \\mu)^2}{2\\sigma^2}\\right).
    $$
    respectively
    $$
    f(y | \\theta_0, \\theta_1) = \\frac{1}{y\\theta_1\\sqrt{2\\pi}}\\exp\\left(-\\frac{(\\log y - \\log \\theta_0)^2}{2\\theta_1^2}\\right).
    $$
    where $y$ is the observed data, $\\mu = \\theta_0$ is the median parameter and $\\sigma = \\theta_1$ is the scale parameter.
    """

    corresponding_gamlss: str = "LOGNO2"
    parameter_names = {0: "mu", 1: "sigma"}
    parameter_support = {
        0: (np.nextafter(0, 1), np.inf),
        1: (np.nextafter(0, 1), np.inf),
    }
    distribution_support = (0, np.inf)
    scipy_dist = st.lognorm
    scipy_names = {"mu": "scale", "sigma": "s"}

    def __init__(
        self,
        loc_link: LinkFunction = Log(),
        scale_link: LinkFunction = Log(),
    ) -> None:
        super().__init__(
            links={
                0: loc_link,
                1: scale_link,
            }
        )

    def theta_to_scipy_params(self, theta: np.ndarray) -> dict:
        mu, sigma = self.theta_to_params(theta)
        return {"s": sigma, "scale": mu, "loc": 0}

    def dl1_dp1(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma = self.theta_to_params(theta)
        if param == 0:
            return (np.log(y) - np.log(mu)) / (mu * sigma**2)
        elif param == 1:
            return ((np.log(y) - np.log(mu)) ** 2 - sigma**2) / (sigma**3)

    def dl2_dp2(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma = self.theta_to_params(theta)
        if param == 0:
            return -1 / (mu**2 * sigma**2)
        elif param == 1:
            return -2 / (sigma**2)

    def dl2_dpp(
        self, y: np.ndarray, theta: np.ndarray, params: Tuple[int, int] = (0, 1)
    ) -> np.ndarray:
        self._validate_dl2_dpp_inputs(y, theta, params)
        return np.zeros_like(y)

    def initial_values(self, y: np.ndarray) -> np.ndarray:
        log_y = np.log(y)
        out = np.zeros((y.shape[0], self.n_params))
        out[:, 0] = np.exp(np.median(log_y, axis=0))
        out[:, 1] = np.std(log_y, axis=0)
        return out



import copy
import numbers
import warnings
from typing import Any, Dict, Literal, Optional, Union

import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin, _fit_context
from sklearn.utils._param_validation import Interval, StrOptions
from sklearn.utils.multiclass import type_of_target
from sklearn.utils.validation import (
    _check_sample_weight,
    check_is_fitted,
    validate_data,
)

from .. import HAS_PANDAS, HAS_POLARS
from ..base import Distribution, EstimationMethod, OndilEstimatorMixin
from ..distributions import Normal
from ..error import OutOfSupportError
from ..gram import init_forget_vector
from ..information_criteria import InformationCriterion
from ..methods import get_estimation_method
from ..scaler import OnlineScaler
from ..utils import calculate_effective_training_length, online_mean_update

if HAS_PANDAS:
    import pandas as pd
if HAS_POLARS:
    import polars as pl  # noqa


class OnlineDistributionalRegression(
    OndilEstimatorMixin, RegressorMixin, BaseEstimator
):
    """The online/incremental GAMLSS class."""

    _parameter_constraints = {
        "forget": [Interval(numbers.Real, 0.0, 1.0, closed="left")],
        "fit_intercept": [bool],
        "scale_inputs": [bool, np.ndarray],
        "regularize_intercept": [bool],
        "method": [EstimationMethod, str],
        "ic": [StrOptions({"aic", "bic", "hqc", "max"})],
        "distribution": [callable],
        "equation": [dict, type(None)],
        "model_selection": [StrOptions({"local_rss", "global_ll"})],
        "prefit_initial": [Interval(numbers.Integral, 0, None, closed="left")],
        "prefit_update": [Interval(numbers.Integral, 0, None, closed="left")],
        "cautious_updates": [bool],
        "max_it_outer": [Interval(numbers.Integral, 1, None, closed="left")],
        "max_it_inner": [Interval(numbers.Integral, 1, None, closed="left")],
        "abs_tol_outer": [Interval(numbers.Real, 0.0, None, closed="left")],
        "abs_tol_inner": [Interval(numbers.Real, 0.0, None, closed="left")],
        "rel_tol_outer": [Interval(numbers.Real, 0.0, None, closed="left")],
        "rel_tol_inner": [Interval(numbers.Real, 0.0, None, closed="left")],
        "step_size": [numbers.Real, dict],
        "verbose": [Interval(numbers.Integral, 0, None, closed="left")],
        "debug": [bool],
        "param_order": [np.ndarray, type(None)],
    }

    def __init__(
        self,
        distribution: Distribution = Normal(),
        equation: Dict[int, Union[str, np.ndarray, list]] = None,
        forget: float | Dict[int, float] = 0.0,
        method: Union[
            str, EstimationMethod, Dict[int, str], Dict[int, EstimationMethod]
        ] = "ols",
        scale_inputs: bool | np.ndarray = True,
        fit_intercept: Union[bool, Dict[int, bool]] = True,
        regularize_intercept: Union[bool, Dict[int, bool]] = False,
        ic: Union[str, Dict] = "aic",
        model_selection: Literal["local_rss", "global_ll"] = "local_rss",
        prefit_initial: int = 0,
        prefit_update: int = 0,
        step_size: float | Dict[int, float] = 1.0,
        verbose: int = 0,
        debug: bool = False,
        param_order: np.ndarray | None = None,
        cautious_updates: bool = False,
        cond_start_val: bool = False,
        max_it_outer: int = 30,
        max_it_inner: int = 30,
        abs_tol_outer: float = 1e-3,
        abs_tol_inner: float = 1e-3,
        rel_tol_outer: float = 1e-5,
        rel_tol_inner: float = 1e-5,
        min_it_outer: int = 1,
    ) -> "OnlineDistributionalRegression":
        """The `OnlineDistributionalRegression()` provides the fit, update and predict methods for linear parametric GAMLSS models.

        For a response variable $Y$ which is distributed according to the distribution $\mathcal{F}(\\theta)$
        with the distribution parameters $\\theta$, we model:

        $$g_k(\\theta_k) = \\eta_k = X_k\\beta_k$$

        where $g_k(\cdot)$ is a link function, which ensures that the predicted distribution parameters are in a
        sensible range (we don't want, e.g. negative standard deviations), and $\eta_k$ is the predictor (on the
        space of the link function). The model is fitted using iterative re-weighted least squares (IRLS).


        !!! note "Tips and Tricks"
            If you're facing issues with non-convergence and/or matrix inversion problems, please enable the `debug` mode and increase the
            logging level by increasing `verbose`.
            In debug mode, the estimator will save the weights, working vectors, derivatives each iteration in a
            according dictionary, i.e. self._debug_weights.
            The keys are composed of a tuple of ints of `(parameter, outer_iteration, inner_iteration)`.
            Very small and/or very large weights (implicitly second derivatives) can be a sign that either start values are not chosen appropriately or
            that the distributional assumption does not fit the data well.

        !!! warning "Debug Mode"
            Please don't use debug more for production models since it saves the `X` matrix and its scaled counterpart, so you will get large
            estimator objects.

        !!! warning "Conditional start values `cond_start_val=False`"
            The `cond_start_val` parameter is considered experimental and may not work as expected.

        !!! warning "Cautious updates `cautious_updates=True`"
            The `cautious_updates` parameter is considered experimental and may not work as expected.

        Args:
            distribution (ondil.Distribution): The parametric distribution to use for modeling the response variable.
            equation (Dict[int, Union[str, np.ndarray, list]], optional): The modeling equation for each distribution parameter. The dictionary should map parameter indices to either the strings `'all'`, `'intercept'`, a numpy array of column indices, or a list of column names. Defaults to None, which uses all covariates for the first parameter and intercepts for others.
            forget (float | Dict[int, float], optional): The forget factor for exponential weighting of past observations. Can be a single float for all parameters or a dictionary mapping parameter indices to floats. Defaults to 0.0.
            method (str | EstimationMethod | Dict[int, str] | Dict[int, EstimationMethod], optional): The estimation method for each parameter. Can be a string, EstimationMethod, or a dictionary mapping parameter indices. Defaults to "ols".
            scale_inputs (bool | np.ndarray, optional): Whether to scale the input features. Can be a boolean or a numpy array specifying scaling per feature. Defaults to True.
            fit_intercept (bool | Dict[int, bool], optional): Whether to fit an intercept for each parameter. Can be a boolean or a dictionary mapping parameter indices. Defaults to True.
            regularize_intercept (bool | Dict[int, bool], optional): Whether to regularize the intercept for each parameter. Can be a boolean or a dictionary mapping parameter indices. Defaults to False.
            ic (str | Dict, optional): Information criterion for model selection (e.g., "aic", "bic"). Can be a string or a dictionary mapping parameter indices. Defaults to "aic".
            model_selection (Literal["local_rss", "global_ll"], optional): Model selection strategy. "local_rss" selects based on local residual sum of squares, "global_ll" uses global log-likelihood. Defaults to "local_rss".
            prefit_initial (int, optional): Number of initial outer iterations with only one inner iteration (for stabilization). Defaults to 0.
            prefit_update (int, optional): Number of initial outer iterations with only one inner iteration during updates. Defaults to 0.
            step_size (float | Dict[int, float], optional): Step size for parameter updates. Can be a float or a dictionary mapping parameter indices. Defaults to 1.0.
            verbose (int, optional): Verbosity level for logging. 0 = silent, 1 = high-level, 2 = per-parameter, 3 = per-iteration. Defaults to 0.
            debug (bool, optional): Enable debug mode. Debug mode will save additional data to the estimator object.
                Currently, we save

                    * self._debug_X_dict
                    * self._debug_X_scaled
                    * self._debug_weights
                    * self._debug_working_vectors
                    * self._debug_dl1dlp1
                    * self._debug_dl2dlp2
                    * self._debug_eta
                    * self._debug_fv
                    * self._debug_coef
                    * self._debug_coef_path

                to the the estimator. Debug mode works in batch and online settings. Note that debug mode is not recommended for production use. Defaults to False.
            param_order (np.ndarray | None, optional): Order in which to fit the distribution parameters. Defaults to None (natural order).
            cautious_updates (bool, optional): If True, use smaller step sizes and more iterations when new data are outliers. Defaults to False.
            cond_start_val (bool, optional): If True, use conditional start values for parameters (experimental). Defaults to False.
            max_it_outer (int, optional): Maximum number of outer iterations for the fitting algorithm. Defaults to 30.
            max_it_inner (int, optional): Maximum number of inner iterations for the fitting algorithm. Defaults to 30.
            abs_tol_outer (float, optional): Absolute tolerance for convergence in the outer loop. Defaults to 1e-3.
            abs_tol_inner (float, optional): Absolute tolerance for convergence in the inner loop. Defaults to 1e-3.
            rel_tol_outer (float, optional): Relative tolerance for convergence in the outer loop. Defaults to 1e-5.
            rel_tol_inner (float, optional): Relative tolerance for convergence in the inner loop. Defaults to 1e-5.
            min_it_outer (int, optional): Minimum number of outer iterations before checking for convergence. Defaults to 1.

        Attributes:
            distribution (Distribution): The distribution used for modeling.
            equation (Dict[int, Union[str, np.ndarray, list]]): The modeling equation for each distribution parameter.
            forget (Dict[int, float]): Forget factor for each distribution parameter.
            fit_intercept (Dict[int, bool]): Whether to fit an intercept for each parameter.
            regularize_intercept (Dict[int, bool]): Whether to regularize the intercept for each parameter.
            ic (Dict[int, str]): Information criterion for model selection for each parameter.
            method (Dict[int, EstimationMethod]): Estimation method for each parameter.
            scale_inputs (bool | np.ndarray): Whether to scale the input features.
            param_order (np.ndarray | None): Order in which to fit the distribution parameters.
            n_observations_ (float): Total number of observations used for fitting.
            n_training_ (Dict[int, int]): Effective training length for each distribution parameter.
            n_features_ (Dict[int, int]): Number of features used for each distribution parameter.
            coef_ (np.ndarray): Coefficients for the fitted model, shape (n_params, n_features).
            coef_path_ (np.ndarray): Coefficients path for the fitted model, shape (n_params, n_iterations, n_features). Only available if `method` is a path-based method like LASSO.

        Returns:
            OnlineDistributionalRegression: The OnlineDistributionalRegression instance.

        """
        self.distribution = distribution
        self.equation = equation
        self.forget = forget
        self.fit_intercept = fit_intercept
        self.regularize_intercept = regularize_intercept
        self.ic = ic
        self.method = method
        self.param_order = param_order
        self.model_selection = model_selection

        # Scaling
        self.scale_inputs = scale_inputs

        # These are global for all distribution parameters
        self.max_it_outer = max_it_outer
        self.max_it_inner = max_it_inner
        self.min_it_outer = min_it_outer
        self.abs_tol_outer = abs_tol_outer
        self.abs_tol_inner = abs_tol_inner
        self.rel_tol_outer = rel_tol_outer
        self.rel_tol_inner = rel_tol_inner

        self.step_size = step_size
        self.cautious_updates = cautious_updates
        self.cond_start_val = cond_start_val
        self.debug = debug
        self.verbose = verbose
        self.prefit_initial = prefit_initial
        self.prefit_update = prefit_update

    @property
    def betas(self):
        warnings.warn(
            "OnlineDistributionalRegression.betas is depreciated in favour of OnlineDistributionalRegression.beta to be consistent with beta_path. "
            "Alternatively, use OnlineDistributionalRegression.coef_ as in sklearn.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.coef_

    @property
    def beta(self):
        check_is_fitted(self)
        return self.coef_

    @property
    def beta_path(self):
        check_is_fitted(self)
        return self.coef_path_

    def _prepare_estimator(self):
        self._equation = self._process_equation(self.equation)
        self._method = {
            p: get_estimation_method(m)
            for p, m in self._process_parameter(
                self.method, default="ols", name="method"
            ).items()
        }
        # Parameters that can be set as dicts or single values need to be processed
        self._forget = self._process_parameter(self.forget, default=0.0, name="forget")
        self._ic = self._process_parameter(self.ic, "aic", name="ic")
        self._fit_intercept = self._process_parameter(
            self.fit_intercept, True, "fit_intercept"
        )
        self._regularize_intercept = self._process_parameter(
            self.regularize_intercept, False, "regularize_intercept"
        )
        self._step_size = self._process_parameter(
            self.step_size, default=1.0, name="step_size"
        )

        if self.param_order is None:
            self._param_order = np.arange(self.distribution.n_params)
        else:
            if all(i in self.param_order for i in range(self.distribution.n_params)):
                self._param_order = self.param_order
            else:
                raise ValueError("All parameters should be in the param_order.")

        self._scaler = OnlineScaler(forget=self._forget[0], to_scale=self.scale_inputs)

        if self.cond_start_val:
            warnings.warn(
                "cond_start_val is considered an experimental feature.",
                UserWarning,
                stacklevel=2,
            )

    def _print_message(self, message, level=0):
        if level <= self.verbose:
            print(f"[{self.__class__.__name__}]", message)

    def _process_parameter(self, attribute: Any, default: Any, name: str) -> None:
        if isinstance(attribute, dict):
            for p in range(self.distribution.n_params):
                if p not in attribute.keys():
                    warnings.warn(
                        f"[{self.__class__.__name__}] "
                        f"No value given for parameter {name} for distribution "
                        f"parameter {p}. Setting default value {default}.",
                        RuntimeWarning,
                        stacklevel=1,
                    )
                    if isinstance(default, dict):
                        attribute[p] = default[p]
                    else:
                        attribute[p] = default
        else:
            # No warning since we expect that floats/strings/ints are either the defaults
            # Or given on purpose for all params the ame
            attribute = {p: attribute for p in range(self.distribution.n_params)}

        return attribute

    def _process_equation(self, equation: Dict):
        """Preprocess the equation object and validate inputs."""
        if equation is None:
            warnings.warn(
                f"[{self.__class__.__name__}] "
                "Equation is not specified. "
                "Per default, will estimate the first distribution parameter by all covariates found in X. "
                "All other distribution parameters will be estimated by an intercept."
            )
            equation = {
                p: "all" if p == 0 else "intercept"
                for p in range(self.distribution.n_params)
            }
        else:
            for p in range(self.distribution.n_params):
                # Check that all distribution parameters are in the equation.
                # If not, add intercept.
                if p not in equation.keys():
                    warnings.warn(
                        f"[{self.__class__.__name__}] "
                        f"Distribution parameter {p} is not in equation. "
                        f"The parameter will be estimated by an intercept."
                    )
                    equation[p] = "intercept"

                if not (
                    isinstance(equation[p], np.ndarray)
                    or (equation[p] in ["all", "intercept"])
                ):
                    if not (
                        isinstance(equation[p], list) and (HAS_PANDAS | HAS_POLARS)
                    ):
                        raise ValueError(
                            "The equation should contain of either: \n"
                            " - a numpy array of dtype int, \n"
                            " - a list of string column names \n"
                            " - or the strings 'all' or 'intercept' \n"
                            f"you have passed {equation[p]} for the distribution parameter {p}."
                        )

        return equation

    @staticmethod
    def _make_intercept(n_observations: int) -> np.ndarray:
        """Make the intercept series as N x 1 array.

        Args:
            y (np.ndarray): Response variable $Y$

        Returns:
            np.ndarray: Intercept array.
        """
        return np.ones((n_observations, 1))

    def _is_intercept_only(self, param: int):
        """Check in the equation whether we model only as intercept"""
        if isinstance(self._equation[param], str):
            return self._equation[param] == "intercept"
        else:
            return False

    def _count_nonzero_coef(self, exclude: int | np.ndarray | None = None) -> int:
        if exclude is None:
            gen = range(self.distribution.n_params)
        else:
            gen = np.delete(np.arange(self.distribution.n_params), exclude)
        return sum(np.count_nonzero(self.coef_[p]) for p in gen)

    def get_J_from_equation(self, X: np.ndarray):
        J = {}
        for p in range(self.distribution.n_params):
            if isinstance(self._equation[p], str):
                if self._equation[p] == "all":
                    J[p] = X.shape[1] + int(self._fit_intercept[p])
                if self._equation[p] == "intercept":
                    J[p] = 1
            elif isinstance(self._equation[p], np.ndarray):
                if np.issubdtype(self._equation[p].dtype, bool):
                    if self._equation[p].shape[0] != X.shape[1]:
                        raise ValueError(f"Shape does not match for param {p}.")
                    J[p] = np.sum(self._equation[p]) + int(self._fit_intercept[p])
                elif np.issubdtype(self._equation[p].dtype, np.integer):
                    if self._equation[p].max() >= X.shape[1]:
                        raise ValueError(f"Shape does not match for param {p}.")
                    J[p] = self._equation[p].shape[0] + int(self._fit_intercept[p])
                else:
                    raise ValueError(
                        "If you pass a np.ndarray in the equation, "
                        "please make sure it is of dtype bool or int."
                    )
            elif isinstance(self._equation[p], list):
                J[p] = len(self._equation[p]) + int(self._fit_intercept[p])
            else:
                raise ValueError("Something unexpected happened")
        return J

    def make_model_array(self, X: Union[np.ndarray], param: int):
        eq = self._equation[param]
        n = X.shape[0]

        # TODO: Check difference between np.array and list more explicitly?
        if isinstance(eq, str) and (eq == "intercept"):
            if not self._fit_intercept[param]:
                raise ValueError(
                    "fit_intercept[param] is false, but equation says intercept."
                )
            out = self._make_intercept(n_observations=n)
        else:
            if isinstance(eq, str) and (eq == "all"):
                if isinstance(X, np.ndarray):
                    out = X
                if HAS_PANDAS and isinstance(X, pd.DataFrame):
                    out = X.to_numpy()
                if HAS_POLARS and isinstance(X, pl.DataFrame):
                    out = X.to_numpy()
            elif isinstance(eq, np.ndarray) | isinstance(eq, list):
                if isinstance(X, np.ndarray):
                    out = X[:, eq]
                if HAS_PANDAS and isinstance(X, pd.DataFrame):
                    out = X.loc[:, eq]
                if HAS_POLARS and isinstance(X, pl.DataFrame):
                    out = X.select(eq).to_numpy()
            else:
                raise ValueError("Did not understand equation. Please check.")

            if self._fit_intercept[param]:
                out = np.hstack((self._make_intercept(n), out))

        return out

    def _fit_select_model(
        self,
        X: np.ndarray,
        y: np.ndarray,
        wv: np.ndarray,
        wt: np.ndarray,
        w: np.ndarray,
        beta_path: np.ndarray,
        param: int,
    ) -> tuple[np.ndarray, np.ndarray, int]:
        """
        Fit and select a model for a specific distribution parameter.

        Selects the best model along a regularization path by evaluating an information criterion
        (e.g., AIC, BIC) based on either local residual sum of squares (RSS) or global log-likelihood (LL).

        Parameters
        ----------
        X : np.ndarray
            The design matrix of shape (n_samples, n_features).
        y : np.ndarray
            The target variable of shape (n_samples,).
        wv : np.ndarray
            Working response vector of shape (n_samples,).
        wt : np.ndarray
            Working weights vector of shape (n_samples,).
        w : np.ndarray
            Sample weights of shape (n_samples,).
        beta_path : np.ndarray
            Array of shape (n_alphas, n_features) containing the coefficient paths for each regularization parameter.
        param : int
            The parameter index for which the model is being selected.

        Returns
        -------
        beta : np.ndarray
            The selected coefficient vector of shape (n_features,) corresponding to the best model.
        model_selection_data : np.ndarray
            The array of RSS or log-likelihood values used for model selection.
        best_ic : int
            The index of the best model according to the information criterion.

        Notes
        -----
        The selection is performed according to the `model_selection` attribute:
        - If "local_rss", the model is selected based on the weighted residual sum of squares.
        - If "global_ll", the model is selected based on the global log-likelihood.
        The function uses an information criterion (e.g., AIC, BIC) to select the best model.
        """
        f = init_forget_vector(self._forget[param], y.shape[0])
        n_nonzero_coef = np.count_nonzero(beta_path, axis=1)
        n_nonzero_coef_other = self._count_nonzero_coef(exclude=param)

        prediction_path = X @ beta_path.T

        if self.model_selection == "local_rss":
            residuals = wv[:, None] - prediction_path
            rss = np.sum(residuals**2 * w[:, None] * wt[:, None] * f[:, None], axis=0)
            rss = rss / np.mean(wt * w * f)
            ic = InformationCriterion(
                n_observations=self.n_training_[param],
                n_parameters=n_nonzero_coef + n_nonzero_coef_other,
                criterion=self._ic[param],
            ).from_rss(rss=rss)
            best_ic = np.argmin(ic)
            beta = beta_path[best_ic, :]
            model_selection_data = rss

        elif self.model_selection == "global_ll":
            ll = np.zeros(self._method[param]._path_length)
            theta = np.copy(self._fv)
            for i in range(self._method[param]._path_length):
                theta[:, param] = self.distribution.link_inverse(
                    prediction_path[:, i], param=param
                )
                ll[i] = np.sum(f * self.distribution.logpdf(y, theta))

            ic = InformationCriterion(
                n_observations=self.n_training_[param],
                n_parameters=n_nonzero_coef + n_nonzero_coef_other,
                criterion=self._ic[param],
            ).from_ll(log_likelihood=ll)
            best_ic = np.argmin(ic)
            beta = beta_path[best_ic, :]
            model_selection_data = ll

        return beta, model_selection_data, best_ic

    def _update_select_model(
        self,
        X: np.ndarray,
        y: np.ndarray,  # observations / response
        wv: np.ndarray,  # working vector
        wt: np.ndarray,  # working weights
        w: np.ndarray,  # sample weights
        beta_path: np.ndarray,
        model_selection_data: Any,
        param: int,
    ):
        """Update and select a model for a specific distribution parameter.

        Parameters
        ----------
        X : np.ndarray
            Feature matrix of shape (n_samples, n_features).
        y : np.ndarray
            Observed response values of shape (n_samples,).
        wv : np.ndarray
            Working vector, typically the adjusted response for IRLS or similar algorithms.
        wt : np.ndarray
            Working weights, used for weighted updates.
        w : np.ndarray
            Sample weights for each observation.
        beta_path : np.ndarray
            Array of candidate coefficient vectors along the regularization path, shape (n_candidates, n_features).
        model_selection_data : Any
            Data carried over from previous model selection steps (e.g., previous RSS or log-likelihood values).
        param : int
            Index of the distribution parameter being updated.
        Returns
        -------
        beta : np.ndarray
            Selected coefficient vector for the best model.
        model_selection_data_new : Any
            Updated model selection data (e.g., RSS or log-likelihood values for the selected model).
        best_ic : int
            Index of the best model according to the information criterion.

        """
        f = init_forget_vector(self._forget[param], y.shape[0])
        n_nonzero_coef = np.count_nonzero(beta_path, axis=1)
        prediction_path = X @ beta_path.T
        if self.model_selection == "local_rss":
            # Denominator
            # TODO: This should go in the online mean update function.
            denom = online_mean_update(
                self._mean_of_weights[param],
                wt,
                self._forget[param],
                self.n_observations_,
            )
            residuals = wv[:, None] - prediction_path
            rss = (
                np.sum((residuals**2) * w[:, None] * wt[:, None] * f[:, None], axis=0)
                + (1 - self._forget[param]) ** y.shape[0]
                * (model_selection_data * self._mean_of_weights[param])
            ) / denom
            n_nonzero_coef_other = self._count_nonzero_coef(exclude=param)
            ic = InformationCriterion(
                n_observations=self.n_training_[param],
                n_parameters=n_nonzero_coef + n_nonzero_coef_other,
                criterion=self._ic[param],
            ).from_rss(rss=rss)
            best_ic = np.argmin(ic)
            model_selection_data_new = rss
        elif self.model_selection == "global_ll":
            ll = np.zeros(self._method[param]._path_length)

            theta = np.copy(self._fv)
            for i in range(self._method[param]._path_length):
                theta[:, param] = self.distribution.link_inverse(
                    prediction_path[:, i], param=param
                )
                ll[i] = np.sum(w * f * self.distribution.logpdf(y, theta))
            ll = ll + (1 - self._forget[param]) ** y.shape[0] * model_selection_data

            ic = InformationCriterion(
                n_observations=self.n_training_[param],
                n_parameters=n_nonzero_coef,
                criterion=self._ic[param],
            ).from_ll(log_likelihood=ll)
            best_ic = np.argmin(ic)
            model_selection_data_new = ll

        beta = beta_path[best_ic, :]
        return beta, model_selection_data_new, best_ic

    def _validate_inputs(self, X: np.ndarray, y: np.ndarray):
        """Validate the input matrices X and y.

        Args:
            X (np.ndarray): Input matrix $X$
            y (np.ndarray): Response vector $y$.

        Raises:
            OutOfSupportError: If the values of $y$ are below the range of the distribution.
            OutOfSupportError: If the values of $y$ are beyond the range of the distribution.
        """
        if np.any(y < self.distribution.distribution_support[0]):
            raise OutOfSupportError(
                message=(
                    "y contains values below the distribution's support. "
                    f"The smallest value in y is {np.min(y)}. "
                    f"The support of the distribution is {str(self.distribution.distribution_support)}."
                )
            )
        if np.any(y > self.distribution.distribution_support[1]):
            raise OutOfSupportError(
                message=(
                    "Y contains values larger than the distribution's support. "
                    f"The smallest value in y is {np.max(y)}. "
                    f"The support of the distribution is {str(self.distribution.distribution_support)}."
                ),
            )

    def _make_iter_schedule(self):
        return np.repeat(self.max_it_inner, repeats=self.max_it_outer)

    def _make_step_size_schedule(self):
        return np.tile(
            A=list(self._step_size.values()),
            reps=(self.max_it_outer, self.max_it_inner, 1),
        ).astype(float)

    @_fit_context(prefer_skip_nested_validation=True)
    def fit(
        self,
        X: np.ndarray,
        y: np.ndarray,
        sample_weight: Optional[np.ndarray] = None,
    ) -> "OnlineDistributionalRegression":
        """Fit the online GAMLSS model.

        This method initializes the model with the given covariate data matrix $X$ and response variable $Y$.

        Args:
            X (np.ndarray): Covariate data matrix $X$.
            y (np.ndarray): Response variable $Y$.
            sample_weight (Optional[np.ndarray], optional): User-defined sample weights. Defaults to None.

        Returns:
            OnlineDistributionalRegression: The fitted OnlineDistributionalRegression instance.

        Raises:
            ValueError: If the equation is not specified correctly.
            OutOfSupportError: If the values of $y$ are below or above the distribution's support.
        """

        self._prepare_estimator()

        # Validate inputs
        X, y = validate_data(self, X=X, y=y, reset=True, dtype=[np.float64, np.float32])
        _ = type_of_target(y, raise_unknown=True)
        sample_weight = _check_sample_weight(X=X, sample_weight=sample_weight)
        self._validate_inputs(X, y)

        self.n_observations_ = sample_weight.sum()
        self.n_training_ = {
            p: calculate_effective_training_length(
                self._forget[p], self.n_observations_
            )
            for p in range(self.distribution.n_params)
        }

        self._fv = self.distribution.initial_values(y=y)
        self.n_features_ = self.get_J_from_equation(X=X)

        # Fit scaler and transform
        self._scaler.fit(X, sample_weight=sample_weight)
        X_scaled = self._scaler.transform(X)
        X_dict = {
            p: self.make_model_array(X_scaled, param=p)
            for p in range(self.distribution.n_params)
        }

        for p, x in X_dict.items():
            cond_num = np.linalg.cond(x)
            if cond_num > 100:
                message = (
                    f"Condition number of X for after scaling (and adding an intercept) for param {p} is {cond_num}. "
                    "This might lead to numerical issues. Consider using a regularized estimation method."
                )
                self._print_message(message=message, level=1)

        if self.debug:
            self._debug_X_dict = X_dict
            self._debug_X_scaled = X_scaled
            self._debug_weights = {}
            self._debug_working_vectors = {}
            self._debug_dl1dlp1 = {}
            self._debug_dl2dlp2 = {}
            self._debug_eta = {}
            self._debug_fv = {}
            self._debug_dv = {}
            self._debug_coef = {}
            self._debug_coef_path = {}

        self._x_gram = {}
        self._y_gram = {}
        self._rss = np.zeros((self.distribution.n_params))
        self._rss_iterations = np.zeros(
            (self.distribution.n_params, self.max_it_outer, self.max_it_inner)
        )

        # For regularized estimation, we have the necessary data to do model selection!
        self._model_selection_data = {}
        self._best_ic = np.zeros(self.distribution.n_params)
        self._best_ic_iterations = np.full(
            (self.distribution.n_params, self.max_it_outer, self.max_it_inner),
            np.nan,
        )

        self._is_regularized = {}
        for p in range(self.distribution.n_params):
            is_regularized = np.repeat(True, self.n_features_[p])
            if self._fit_intercept[p] and not (
                self._regularize_intercept[p] | self._is_intercept_only(p)
            ):
                is_regularized[0] = False
            self._is_regularized[p] = is_regularized

        self.coef_ = {
            p: np.zeros(self.n_features_[p]) for p in range(self.distribution.n_params)
        }
        self.coef_path_ = {p: None for p in range(self.distribution.n_params)}

        # We need to track the sum of weights for each
        # distribution parameter for online model selection
        self._sum_of_weights = {}
        self._mean_of_weights = {}

        self._schedule_iteration = self._make_iter_schedule()
        self._schedule_step_size = self._make_step_size_schedule()

        self._it_inner = np.zeros(self.distribution.n_params)
        if self.prefit_initial > 0:
            message = (
                f"Setting max_it_inner to {self.prefit_initial} for first iteration"
            )
            self._print_message(message=message, level=1)
            self._schedule_iteration[: self.prefit_initial] = 1
            self._current_min_it_outer = int(self.prefit_initial)
        else:
            self._current_min_it_outer = int(self.min_it_outer)

        message = "Starting fit call"
        self._print_message(message=message, level=1)
        (
            self._global_dev,
            self._it_outer,
        ) = self._outer_fit(
            X=X_dict,
            y=y,
            w=sample_weight,
        )
        message = "Finished fit call"
        self._print_message(message=message, level=1)
        return self

    @_fit_context(prefer_skip_nested_validation=True)
    def update(
        self,
        X: np.ndarray,
        y: np.ndarray,
        sample_weight: Optional[np.ndarray] = None,
    ):
        """Update the fit for the online GAMLSS Model.

        Args:
            X (np.ndarray): Covariate data matrix $X$.
            y (np.ndarray): Response variable $Y$.
            sample_weight (Optional[np.ndarray], optional): User-defined sample weights. Defaults to None (all observations have the same weight).
        """
        X, y = validate_data(
            self, X=X, y=y, reset=False, dtype=[np.float64, np.float32]
        )
        sample_weight = _check_sample_weight(X=X, sample_weight=sample_weight)
        _ = type_of_target(y, raise_unknown=True)
        self._validate_inputs(X, y)

        self.n_observations_ += sample_weight.sum()
        self.n_training_ = {
            p: calculate_effective_training_length(
                self._forget[p], self.n_observations_
            )
            for p in range(self.distribution.n_params)
        }

        self._fv = self.predict_distribution_parameters(X, what="response")

        self._scaler.update(X, sample_weight=sample_weight)
        X_scaled = self._scaler.transform(X)
        X_dict = {
            p: self.make_model_array(X_scaled, param=p)
            for p in range(self.distribution.n_params)
        }

        if self.debug:
            self._debug_X_dict = X_dict
            self._debug_X_scaled = X_scaled
            self._debug_weights = {}
            self._debug_working_vectors = {}
            self._debug_dl1dlp1 = {}
            self._debug_dl2dlp2 = {}
            self._debug_eta = {}
            self._debug_fv = {}
            self._debug_coef = {}
            self._debug_coef_path = {}

        ## Reset rss and ic to avoid confusion
        ## These are only for viewing, not read!!
        self._best_ic[:] = 0
        self._best_ic_iterations[:] = 0
        self._rss_iterations[:] = 0

        ## Copy old values for updating
        self._x_gram_new = copy.copy(self._x_gram)
        self._y_gram_new = copy.copy(self._y_gram)
        self._sum_of_weights_new = copy.copy(self._sum_of_weights)
        self._mean_of_weights_new = copy.copy(self._mean_of_weights)

        self._rss_old = copy.copy(self._rss)
        self._model_selection_data_old = copy.copy(self._model_selection_data)

        # Clean schedules for iterations and step size
        self._schedule_iteration = self._make_iter_schedule()
        self._schedule_step_size = self._make_step_size_schedule()

        if self.prefit_update > 0:
            message = (
                f"Setting max_it_inner to {self.prefit_update} for first iteration"
            )
            self._print_message(message=message, level=1)
            self._schedule_iteration[: self.prefit_update] = 1
            self._current_min_it_outer = int(self.prefit_update)

        if self.cautious_updates:
            lower = self.distribution.quantile(0.005, self._fv).item()
            upper = self.distribution.quantile(0.995, self._fv).item()

            if np.any(y < lower) or np.any(y > upper):
                message = (
                    f"New observations are outliers given current estimates. "
                    f"y: {y}, Lower bound: {lower}, upper bound: {upper}"
                )
                self._print_message(message=message, level=1)
                self._schedule_iteration[:4] = 1
                self._schedule_step_size[:4, :, :] = 0.25
                self._current_min_it_outer = 4
            else:
                self._current_min_it_outer = int(self.min_it_outer)

        message = "Starting update call"
        self._print_message(message=message, level=1)
        (
            self._global_dev,
            self._it_outer,
        ) = self._outer_update(
            X=X_dict,
            y=y,
            w=sample_weight,
        )

        self._x_gram = copy.copy(self._x_gram_new)
        self._y_gram = copy.copy(self._y_gram_new)
        self._sum_of_weights = copy.copy(self._sum_of_weights_new)
        self._mean_of_weights = copy.copy(self._mean_of_weights_new)
        message = "Finished update call"
        self._print_message(message=message, level=1)
        return self

    def _outer_update(self, X, y, w):
        ## for new observations:
        global_di = np.sum(-2 * self.distribution.logpdf(y, self._fv) * w)
        global_dev = (1 - self._forget[0]) ** y.shape[0] * self._global_dev + global_di
        global_dev_old = global_dev + 1000
        it_outer = 0

        while True:
            # Check relative congergence
            if it_outer >= self._current_min_it_outer:
                if (
                    np.abs(global_dev_old - global_dev) / np.abs(global_dev_old)
                    < self.rel_tol_outer
                ):
                    break

                if np.abs(global_dev_old - global_dev) < self.abs_tol_outer:
                    break

                # if global_dev > global_dev_old:
                #     message = (
                #         f"Outer iteration {it_outer}: Global deviance increased. Breaking."
                #         f"Current LL {global_dev}, old LL {global_dev_old}"
                #     )
                #     self._print_message(message=message, level=1)
                #     break

                if it_outer >= self.max_it_outer:
                    break

            global_dev_old = global_dev
            it_outer += 1

            for param in self._param_order:
                global_dev = self._inner_update(
                    X=X,
                    y=y,
                    w=w,
                    it_outer=it_outer,
                    param=param,
                    dv=global_dev,
                )
                message = f"Outer iteration {it_outer}: Fitted param {param}: Current deviance {global_dev}"
                self._print_message(message=message, level=2)

            message = (
                f"Outer iteration {it_outer}: Finished: current deviance {global_dev}"
            )
            self._print_message(message=message, level=1)

        return global_dev, it_outer

    def _outer_fit(self, X, y, w):
        global_di = -2 * self.distribution.logpdf(y, self._fv)
        global_dev = np.sum(w * global_di)
        global_dev_old = global_dev + 1000
        it_outer = 0

        while True:
            # Check relative congergence
            if it_outer >= self._current_min_it_outer:
                if (
                    np.abs(global_dev_old - global_dev) / np.abs(global_dev_old)
                    < self.rel_tol_outer
                ):
                    break

                if np.abs(global_dev_old - global_dev) < self.abs_tol_outer:
                    break

                if it_outer >= self.max_it_outer:
                    break

                if global_dev > global_dev_old:
                    message = (
                        f"Outer iteration {it_outer}: Global deviance increased. Breaking."
                        f"Current deviance {global_dev}, old deviance {global_dev_old}"
                    )
                    self._print_message(message=message, level=0)
                    break

            global_dev_old = float(global_dev)
            it_outer += 1

            for param in self._param_order:
                global_dev = self._inner_fit(
                    X=X,
                    y=y,
                    w=w,
                    param=param,
                    it_outer=it_outer,
                    dv=global_dev,
                )
                message = f"Outer iteration {it_outer}: Fitted param {param}: current deviance {global_dev}"
                self._print_message(message=message, level=2)

            message = f"Outer iteration {it_outer}: Finished. Current deviance {global_dev}, old deviance {global_dev_old}"
            self._print_message(message=message, level=1)

        return (global_dev, it_outer)

    def _inner_fit(
        self,
        X,
        y,
        w,
        it_outer,
        param,
        dv,
    ):
        dv_start = np.sum(-2 * self.distribution.logpdf(y, self._fv) * w)
        dv_iterations = np.repeat(dv_start, self.max_it_inner + 1)
        fv_it = copy.copy(self._fv)
        fv_it_new = copy.copy(self._fv)
        step_decrease_counter = 0
        terminate = False
        bad_state = False

        message = f"Starting inner iteration param {param}, outer iteration {it_outer}, start DV {dv_start}"
        self._print_message(message=message, level=2)

        for it_inner in range(self._schedule_iteration[it_outer - 1]):
            # We can improve the fit by taking the conditional
            # start values for the first outer iteration and the first inner iteration
            # as soon the first parameter is fitted.
            step_it = self._schedule_step_size[it_outer - 1, it_inner, param]

            if (
                (it_inner == 0)
                and (it_outer == 1)
                and (param >= 1)
                and self.cond_start_val
            ):
                fv_it = self.distribution.calculate_conditional_initial_values(
                    y=y,
                    theta=fv_it,
                    param=param,
                )
                message = f"Outer iteration {it_outer}: Fitting Parameter {param}: Using conditional start value {fv_it[0, param]}."
                self._print_message(message=message, level=3)

            eta = self.distribution.link_function(fv_it[:, param], param=param)
            dr = 1 / self.distribution.link_inverse_derivative(eta, param=param)
            dl1dp1 = self.distribution.dl1_dp1(y, fv_it, param=param)
            dl2dp2 = self.distribution.dl2_dp2(y, fv_it, param=param)
            wt = -(dl2dp2 / (dr * dr))
            wt = np.clip(wt, 1e-10, 1e10)
            wv = eta + dl1dp1 / (dr * wt)

            if self.debug:
                key = (param, it_outer, it_inner)
                self._debug_weights[key] = copy.copy(wt)
                self._debug_working_vectors[key] = copy.copy(wv)
                self._debug_dl1dlp1[key] = copy.copy(dl1dp1)
                self._debug_dl2dlp2[key] = copy.copy(dl2dp2)
                self._debug_eta[key] = copy.copy(eta)
                self._debug_fv[key] = copy.copy(fv_it)
                self._debug_dv[key] = float(dv_iterations[it_inner])

            ## Create the X and Y Gramian and the weight
            x_gram_it = self._method[param].init_x_gram(
                X=X[param],
                weights=(w * wt),
                forget=self._forget[param],
            )
            y_gram_it = self._method[param].init_y_gram(
                X=X[param],
                y=wv,
                weights=(w * wt),
                forget=self._forget[param],
            )
            # Select the model if we have a path-based method
            if self._method[param]._path_based_method:
                if (it_inner == 0) & (it_outer == 1):
                    beta_path_it = self._method[param].fit_beta_path(
                        x_gram=x_gram_it,
                        y_gram=y_gram_it,
                        is_regularized=self._is_regularized[param],
                    )
                elif it_inner > 0:
                    beta_path_it = self._method[param].update_beta_path(
                        x_gram=x_gram_it,
                        y_gram=y_gram_it,
                        is_regularized=self._is_regularized[param],
                        beta_path=beta_path_it,
                    )
                elif (it_outer > 1) & (it_inner == 0):
                    beta_path_it = self._method[param].update_beta_path(
                        x_gram=x_gram_it,
                        y_gram=y_gram_it,
                        is_regularized=self._is_regularized[param],
                        beta_path=self.coef_path_[param],
                    )
                else:
                    print("This should not happen")

                beta_it, model_selection_data_it, best_ic_it = self._fit_select_model(
                    X=X[param],
                    y=y,
                    w=w,
                    wv=wv,
                    wt=wt,
                    beta_path=beta_path_it,
                    param=param,
                )
                self._best_ic_iterations[param, it_outer - 1, it_inner] = best_ic_it

            else:
                beta_path_it = None
                beta_it = self._method[param].fit_beta(
                    x_gram=x_gram_it,
                    y_gram=y_gram_it,
                    is_regularized=self._is_regularized[param],
                )

            if self.debug:
                self._debug_coef[key] = copy.copy(beta_it)
                self._debug_coef_path[key] = copy.copy(beta_path_it)

            # Calculate the prediction, residuals and RSS
            f = init_forget_vector(self._forget[param], y.shape[0])
            prediction_it = step_it * (X[param] @ beta_it.T) + (1 - step_it) * eta
            residuals_it = wv - prediction_it
            rss_it = np.sum(residuals_it**2 * wt * w * f) / np.mean(wt * w * f)

            # Calculate the fitted values and the deviance
            fv_it_new[:, param] = self.distribution.link_inverse(
                prediction_it, param=param
            )
            dv_it = np.sum(-2 * self.distribution.logpdf(y, fv_it_new) * w)
            dv_old = dv_iterations[it_inner]
            dv_increasing = dv_it > dv_old

            # This should really not happen unless your start values are
            # way to good and the model cannot reach these
            bad_state = dv_it > dv_iterations[0]

            # print(dv_it, dv_old, dv_increasing)
            message = f"Outer iteration {it_outer}: Fitting Parameter {param}: Inner iteration {it_inner}: Current Deviance {dv_it}"
            self._print_message(message=message, level=3)

            if dv_increasing and it_inner < (
                self._schedule_iteration[it_outer - 1] - 1
            ):
                # print("Blabal")
                step_decrease_counter += 1
                self._schedule_step_size[it_outer - 1, it_inner + 1, param] = (
                    step_it / 2
                )
                self._print_message(
                    f"Deviance increasing, step size halved. {step_decrease_counter}",
                    level=1,
                )
                if step_decrease_counter > 5:
                    message = f"Step size too small. Parameter {param}, Outer iteration {it_outer}, Inner iteration {it_inner}."
                    self._print_message(message=message, level=1)
                    terminate = True

            if (it_outer == 1) & (it_inner >= 1) | (it_outer >= 2):
                # Allow to break in principle.
                if abs(dv_old - dv_it) <= self.abs_tol_inner:
                    terminate = True
                if abs(dv_old - dv_it) / abs(dv_old) < self.rel_tol_inner:
                    terminate = True
                if it_inner == (self.max_it_inner - 1):
                    message = f"Reached max inner iteration in inner fit. Parameter:{param}, Outer iteration: {it_outer}, Inner iteration: {it_inner}."
                    self._print_message(message=message, level=3)
                    terminate = True

            if terminate and (it_outer == 1) and bad_state:
                message = (
                    f"The model ended in a bad state in the first outer iteration of param {param}. This is not a good sign.  \n"
                    f"The deviance increased from the start values to the current fit in inner iteration {it_inner}. \n"
                    f"The deviance is {dv_it} and the starting deviance for this inner iterations is {dv_iterations[0]}. \n"
                    "Please check your data and model. \n"
                    "Please turn on logging (verbose=3, debug=True) and check the debug information. \n"
                    "Consider using a pre-fit via the iteration_schedule and set the inner iterations to 1-2 for the first outer iteration."
                )
                self._print_message(message=message, level=0)

            if terminate:
                break
            if not dv_increasing:
                # print("deviance decreasing, write to fv_it")
                fv_it[:, param] = fv_it_new[:, param]
                # Set the deviance for the next inner iteration
                dv_iterations[(it_inner + 1) :] = dv_it

        if (not bad_state) or (it_outer == 1):
            # Write everything to the class
            self._x_gram[param] = x_gram_it
            self._y_gram[param] = y_gram_it
            self.coef_[param] = beta_it
            self.coef_path_[param] = beta_path_it
            self._fv[:, param] = fv_it_new[:, param]

            # Sum and mean of the weights
            self._sum_of_weights[param] = np.sum(w * wt)
            self._mean_of_weights[param] = np.mean(w * wt)

            # RSS
            self._rss[param] = rss_it
            self._rss_iterations[param, it_outer - 1, it_inner - 1] = rss_it

            if self._method[param]._path_based_method:
                self._model_selection_data[param] = model_selection_data_it
                self._best_ic[param] = best_ic_it

        self._it_inner[param] = it_inner
        return dv_it

    def _inner_update(
        self,
        X,
        y,
        w,
        it_outer,
        dv,
        param,
    ):
        # di = -2 * self.distribution.logpdf(y, self._fv )
        # dv = (1 - self._forget[0]) * self._global_dev + np.sum(di * w)
        # olddv = dv + 1

        dv_start = (
            np.sum(-2 * self.distribution.logpdf(y, self._fv) * w)
            + (1 - self._forget[0]) ** y.shape[0]
            * self._global_dev  # global dev is previous observation / fit
        )
        dv_iterations = np.repeat(dv_start, self.max_it_inner + 1)
        fv_it = copy.copy(self._fv)
        fv_it_new = copy.copy(self._fv)
        step_decrease_counter = 0
        terminate = False

        for it_inner in range(self._schedule_iteration[it_outer - 1]):
            step_it = self._schedule_step_size[it_outer - 1, it_inner, param]

            eta = self.distribution.link_function(fv_it[:, param], param=param)
            dr = 1 / self.distribution.link_inverse_derivative(eta, param=param)
            dl1dp1 = self.distribution.dl1_dp1(y, fv_it, param=param)
            dl2dp2 = self.distribution.dl2_dp2(y, fv_it, param=param)
            wt = -(dl2dp2 / (dr * dr))
            wt = np.clip(wt, -1e10, 1e10)
            wv = eta + dl1dp1 / (dr * wt)

            if self.debug:
                key = (param, it_outer, it_inner)
                self._debug_weights[key] = copy.copy(wt)
                self._debug_working_vectors[key] = copy.copy(wv)
                self._debug_dl1dlp1[key] = copy.copy(dl1dp1)
                self._debug_dl2dlp2[key] = copy.copy(dl2dp2)
                self._debug_eta[key] = copy.copy(eta)
                self._debug_fv[key] = copy.copy(fv_it)

            x_gram_it = self._method[param].update_x_gram(
                gram=self._x_gram[param],
                X=X[param],
                weights=(w * wt),
                forget=self._forget[param],
            )
            y_gram_it = self._method[param].update_y_gram(
                gram=self._y_gram[param],
                X=X[param],
                y=wv,
                weights=(w * wt),
                forget=self._forget[param],
            )
            # Select the model if we have a path-based method
            if self._method[param]._path_based_method:
                beta_path_it = self._method[param].update_beta_path(
                    x_gram=x_gram_it,
                    y_gram=y_gram_it,
                    beta_path=self.coef_path_[param],
                    is_regularized=self._is_regularized[param],
                )

                beta_it, model_selection_data_it, best_ic_it = (
                    self._update_select_model(
                        X=X[param],
                        y=y,
                        w=w,
                        wv=wv,
                        wt=wt,
                        beta_path=beta_path_it,
                        model_selection_data=self._model_selection_data_old[param],
                        param=param,
                    )
                )
                self._best_ic[param] = best_ic_it
                self._best_ic_iterations[param, it_outer - 1, it_inner] = best_ic_it
            else:
                beta_it = self._method[param].update_beta(
                    x_gram=x_gram_it,
                    y_gram=y_gram_it,
                    beta=self.coef_[param],
                    is_regularized=self._is_regularized[param],
                )
                beta_path_it = None
                model_selection_data_it = None

            if self.debug:
                self._debug_coef[key] = copy.copy(beta_it)
                self._debug_coef_path[key] = copy.copy(beta_path_it)

            # Calculate the prediction, residuals and RSS
            f = init_forget_vector(self._forget[param], y.shape[0])
            prediction_it = X[param] @ beta_it.T
            residuals_it = wv - prediction_it
            rss_it = np.sum(residuals_it**2 * wt * w * f) / np.mean(wt * w * f)

            denom = online_mean_update(
                self._mean_of_weights[param],
                wt,
                self._forget[param],
                self.n_observations_,
            )
            sum_of_rss_it = (
                rss_it
                + (1 - self._forget[param]) ** y.shape[0]
                * (self._rss_old[param] * self._mean_of_weights[param])
            ) / denom

            # Calculate the fitted values and the deviance
            fv_it_new[:, param] = self.distribution.link_inverse(
                prediction_it, param=param
            )
            dv_it = (
                np.sum(-2 * self.distribution.logpdf(y, fv_it_new) * w)
                + (1 - self._forget[0]) ** y.shape[0] * self._global_dev
            )
            dv_old = dv_iterations[it_inner]
            dv_increasing = dv_it > dv_old

            if dv_increasing:
                step_decrease_counter += 1
                self._schedule_step_size[it_outer - 1, it_inner, param] = step_it / 2
                self._print_message(
                    f"Deviance increasing, step size halved. {step_decrease_counter}",
                    level=1,
                )
                if step_decrease_counter > 5:
                    message = f"Step size too small. Parameter {param}, Outer iteration {it_outer}, Inner iteration {it_inner}."
                    self._print_message(message=message, level=1)
                    terminate = True

            if (not dv_increasing) | (it_inner < self.max_it_inner - 1):
                # Allow to break in principle.
                if abs(dv_old - dv_it) <= self.abs_tol_inner:
                    terminate = True
                if abs(dv_old - dv_it) / abs(dv_old) < self.rel_tol_inner:
                    terminate = True
                if it_inner == (self.max_it_inner - 1):
                    message = f"Reached max inner iteration in inner fit. Parameter:{param}, Outer iteration: {it_outer}, Inner iteration: {it_inner}."
                    self._print_message(message=message, level=3)
                    terminate = True

            message = f"Outer iteration {it_outer}: Fitting Parameter {param}: Inner iteration {it_inner}: Current DV {dv_it}, previous DV {dv_old}, step size {step_it}"
            self._print_message(message=message, level=3)

            if terminate:
                break
            if not dv_increasing:
                # print("deviance decreasing, write to fv_it")
                fv_it[:, param] = fv_it_new[:, param]
                # Set the deviance for the next inner iteration
                dv_iterations[(it_inner + 1) :] = dv_it

            # olddv = dv
            # di = -2 * self.distribution.logpdf(y, self._fv )
            # dv = np.sum(di * w)

        # Assign to class variables
        self._x_gram_new[param] = x_gram_it
        self._y_gram_new[param] = y_gram_it
        self._fv[:, param] = fv_it_new[:, param]

        self.coef_[param] = beta_it
        self.coef_path_[param] = beta_path_it
        self._rss[param] = sum_of_rss_it
        self._rss_iterations[param, it_outer - 1, it_inner] = sum_of_rss_it
        self._model_selection_data[param] = model_selection_data_it

        # Update the weights
        self._sum_of_weights_new[param] = (
            np.sum(w * wt) + (1 - self._forget[param]) * self._sum_of_weights[param]
        )
        self._mean_of_weights_new[param] = (
            self._sum_of_weights_new[param] / self.n_training_[param]
        )
        self._it_inner[param] = it_inner
        return dv_it

    def predict(self, X: np.ndarray) -> np.ndarray:
        """Predict the mean of the response distribution.

        Args:
            X (np.ndarray): Covariate matrix $X$. Shape should be (n_samples, n_features).
        Raises:
            NotFittedError: If the model is not fitted yet.

        Returns:
            Predictions (np.ndarray): Predictions
        """
        theta = self.predict_distribution_parameters(X)
        return self.distribution.mean(theta)

    def predict_median(self, X: np.ndarray):
        """Predict the median of the distribution.

        Args:
            X (np.ndarray): Covariate matrix $X$. Shape should be (n_samples, n_features).
        Raises:
            NotFittedError: If the model is not fitted yet.

        Returns:
            Predictions (np.ndarray): Predicted median of the distribution. Shape will be (n_samples,).
        """
        theta = self.predict_distribution_parameters(X)
        return self.distribution.median(theta)

    def predict_distribution_parameters(
        self,
        X: np.ndarray,
        what: str = "response",
        return_contributions: bool = False,
    ) -> np.ndarray:
        """Predict the distibution parameters given input data.

        Args:
            X (np.ndarray): Design matrix.
            what (str, optional): Predict the response or the link. Defaults to "response". Remember the  GAMLSS models $g(\\theta) = X^T\\beta$. Predict `"link"` will output $X^T\\beta$, predict `"response"` will output $g^{-1}(X^T\\beta)$. Usually, you want predict = `"response"`.
            return_contributions (bool, optional): Whether to return a `Tuple[prediction, contributions]` where the contributions of the individual covariates for each distribution parameter's predicted value is specified. Defaults to False.

        Raises:
            ValueError: Raises if `what` is not in `["link", "response"]`.

        Returns:
            Predictions (np.ndarray): Predicted values for the distribution of shape (n_samples, n_params) where n_params is the number of distribution parameters.
        """
        check_is_fitted(self)
        X = validate_data(self, X=X, reset=False, dtype=[np.float64, np.float32])

        X_scaled = self._scaler.transform(X=X)
        X_dict = {
            p: self.make_model_array(X_scaled, p)
            for p in range(self.distribution.n_params)
        }
        prediction = [x @ b.T for x, b in zip(X_dict.values(), self.coef_.values())]

        if return_contributions:
            contribution = [
                x * b.T for x, b in zip(X_dict.values(), self.coef_.values())
            ]

        if what == "response":
            prediction = [
                self.distribution.link_inverse(p, param=i)
                for i, p in enumerate(prediction)
            ]
        elif what == "link":
            pass
        else:
            raise ValueError("Should be 'response' or 'link'.")

        prediction = np.stack(prediction, axis=1)

        if return_contributions:
            return (prediction, contribution)
        else:
            return prediction

    def predict_quantile(
        self,
        X: np.ndarray,
        quantile: float | np.ndarray,
    ) -> np.ndarray:
        """Predict the quantile(s) of the distribution.

        Args:
            X (np.ndarray): Covariate matrix $X$. Shape should be (n_samples, n_features).
            quantile (float | np.ndarray): Quantile(s) to predict.

        Returns:
            np.ndarray: Predicted quantile(s) of the distribution. Shape will be (n_samples, n_quantiles).
        """
        check_is_fitted(self)
        X = validate_data(self, X=X, reset=False, dtype=[np.float64, np.float32])

        theta = self.predict_distribution_parameters(X, what="response")
        if isinstance(quantile, np.ndarray):
            quantile_pred = self.distribution.ppf(quantile[:, None], theta).T
        else:
            quantile_pred = self.distribution.ppf(quantile, theta).reshape(-1, 1)

        return quantile_pred

    def get_debug_information(
        self,
        variable: str = "coef",
        param: int = 0,
        it_outer: int = 1,
        it_inner: int = 1,
    ):
        """Get debug information for a specific variable, parameter, outer iteration and inner iteration.

        We currently support the following variables:

        * "X_dict": The design matrix for the distribution parameter.
        * "X_scaled": The scaled design matrix.
        * "weights": The sample weights for the distribution parameter.
        * "working_vectors": The working vectors for the distribution parameter.
        * "dl1dlp1": The first derivative of the log-likelihood with respect to the distribution parameter.
        * "dl2dlp2": The second derivative of the log-likelihood with respect to the distribution parameter.
        * "eta": The linear predictor for the distribution parameter.
        * "fv": The fitted values for the distribution parameter.
        * "dv": The deviance for the distribution parameter.
        * "coef": The coefficients for the distribution parameter.
        * "coef_path": The coefficients path for the distribution parameter.

        Args:
            variable (str): The variable to get debug information for. Defaults to "coef".
            param (int): The distribution parameter to get debug information for. Defaults to 0.
            it_outer (int): The outer iteration to get debug information for. Defaults to 1.
            it_inner (int): The inner iteration to get debug information for. Defaults to 1.
        Returns:
            Any: The debug information for the specified variable, parameter, outer iteration and inner iteration.
        Raises:
            ValueError: If debug mode is not enabled.

        """
        if not self.debug:
            raise ValueError(
                "Debug mode is not enabled. Please set debug=True when initializing the OnlineDistributionalRegression estimator."
            )

        key = (param, it_outer, it_inner + 1)
        return getattr(self, f"_debug_{variable}")[key]



import ondil.estimators
import pytest
from sklearn.utils.estimator_checks import check_estimator

import ondil

EXPECTED_FAILED_CHECKS = {
    "check_sample_weight_equivalence_on_dense_data": "To few data points to test this check in the original sklaern implementation.",
}

# https://scikit-learn.org/stable/modules/generated/sklearn.utils.estimator_checks.check_estimator.html#sklearn.utils.estimator_checks.check_estimator


@pytest.mark.parametrize("scale_inputs", [False, True])
@pytest.mark.parametrize("fit_intercept", [False, True])
@pytest.mark.parametrize("method", ["ols", "lasso", "elasticnet"])
@pytest.mark.parametrize("ic", ["aic", "bic", "hqc", "max"])
def test_sklearn_compliance_linear_model(scale_inputs, fit_intercept, method, ic):
    estimator = ondil.estimators.OnlineLinearModel(
        fit_intercept=fit_intercept,
        scale_inputs=scale_inputs,
        method=method,
        ic=ic,
    )
    check_estimator(
        estimator, on_fail="raise", expected_failed_checks=EXPECTED_FAILED_CHECKS
    )


@pytest.mark.parametrize("scale_inputs", [False, True])
@pytest.mark.parametrize("method", ["ols", "lasso", "elasticnet"])
@pytest.mark.parametrize("ic", ["aic", "bic", "hqc", "max"])
def test_sklearn_compliance_online_gamlss(scale_inputs, method, ic):
    estimator = ondil.estimators.OnlineDistributionalRegression(
        scale_inputs=scale_inputs,
        method=method,
        ic=ic,
    )
    check_estimator(
        estimator, on_fail="raise", expected_failed_checks=EXPECTED_FAILED_CHECKS
    )


def test_sklearn_compliance_scaler():
    estimator = ondil.OnlineScaler()
    check_estimator(
        estimator,
        on_fail="raise",
    )



from typing import Literal, Optional

import numpy as np

from ..base import EstimationMethod
from ..coordinate_descent import online_coordinate_descent
from ..gram import init_gram, init_y_gram, update_gram, update_y_gram


class Ridge(EstimationMethod):
    """Single-lambda Ridge Estimation.

    The ridge method runs coordinate descent for a single lambda.

    We allow to pass user-defined lower and upper bounds for the coefficients.
    The coefficient bounds must be an `numpy` array of the length of `X` respectively of the number of variables in the
    equation _plus the intercept, if you fit one_. This allows to box-constrain the coefficients to a certain range.

    Lastly, we have some rather technical parameters like the number of coordinate descent iterations,
    whether you want to cycle randomly and for which tolerance you want to break. We use active set iterations, i.e.
    after the first coordinate-wise update for each regularization strength, only non-zero coefficients are updated.

    We use `numba` to speed up the coordinate descent algorithm.
    """

    def __init__(
        self,
        lambda_reg: float | None = None,
        start_beta: np.ndarray | None = None,
        selection: Literal["cyclic", "random"] = "cyclic",
        beta_lower_bound: np.ndarray | None = None,
        beta_upper_bound: np.ndarray | None = None,
        tolerance: float = 1e-4,
        max_iterations: int = 1000,
    ):
        """
        Initializes the Ridge method with the specified parameters.

        Args:
            lambda_reg (float): Regularization parameter. Must be greater than 0. Higher values lead to more regularization. If not set, the average variance of the features is used as the default.
            selection (Literal["cyclic", "random"]): Method to select features during the path. Default is "cyclic".
            beta_lower_bound (np.ndarray | None): Lower bound for the coefficients. Default is None.
            beta_upper_bound (np.ndarray | None): Upper bound for the coefficients. Default is None.
            tolerance (float): Tolerance for the optimization. Default is 1e-4.
            max_iterations (int): Maximum number of iterations for the optimization. Default is 1000.
        """
        super().__init__(
            _path_based_method=False,
            _accepts_bounds=True,
            _accepts_selection=False,
        )
        self.beta_lower_bound = beta_lower_bound
        self.beta_upper_bound = beta_upper_bound
        self.lambda_reg = lambda_reg
        self.selection = selection
        self.tolerance = tolerance
        self.max_iterations = max_iterations
        self.start_beta = start_beta

    def _set_and_validate_bounds(self, x_gram: np.ndarray) -> None:
        J = x_gram.shape[1]
        if self.beta_lower_bound is None:
            self.beta_lower_bound = np.repeat(-np.inf, J)
        if self.beta_upper_bound is None:
            self.beta_upper_bound = np.repeat(np.inf, J)

        if len(self.beta_lower_bound) != J:
            raise ValueError("Lower bound does not have correct length")
        if len(self.beta_upper_bound) != J:
            raise ValueError("Upper bound does not have correct length")

    @staticmethod
    def init_x_gram(X, weights, forget):
        return init_gram(X=X, w=weights, forget=forget)

    @staticmethod
    def init_y_gram(X, y, weights, forget):
        return init_y_gram(X, y, w=weights, forget=forget)

    @staticmethod
    def update_x_gram(gram, X, weights, forget):
        return update_gram(gram, X, w=weights, forget=forget)

    @staticmethod
    def update_y_gram(gram, X, y, weights, forget):
        return update_y_gram(gram, X, y, forget=forget, w=weights)

    def fit_beta(self, x_gram, y_gram, is_regularized):
        self._set_and_validate_bounds(x_gram=x_gram)

        # beta = (x_gram @ y_gram).squeeze(-1)

        if self.start_beta is not None:
            # Use user-defined start value
            beta = self.start_beta
        else:
            # Use OLS solution for initialization if x_gram is invertible
            if np.linalg.matrix_rank(x_gram) == x_gram.shape[0]:
                beta = (np.linalg.inv(x_gram) @ y_gram).squeeze(-1)
            else:
                # If x_gram is not invertible, initialize with zeros
                beta = np.zeros(x_gram.shape[1])

        if self.lambda_reg is None:
            self.lambda_reg = np.linalg.trace(x_gram) / x_gram.shape[0]

        # Ensure that beta is a column vector
        if beta.ndim > 1:
            raise ValueError(
                "The beta vector must be a column vector. Please check the input data."
            )

        beta, _ = online_coordinate_descent(
            x_gram=x_gram,
            y_gram=y_gram.squeeze(-1),
            beta=beta,
            regularization=self.lambda_reg,
            is_regularized=is_regularized,
            alpha=0.0,
            beta_lower_bound=self.beta_lower_bound,
            beta_upper_bound=self.beta_upper_bound,
            selection=self.selection,
            tolerance=self.tolerance,
            max_iterations=self.max_iterations,
        )
        return beta

    def update_beta(self, x_gram, y_gram, beta, is_regularized):
        beta, _ = online_coordinate_descent(
            x_gram=x_gram,
            y_gram=y_gram.squeeze(-1),
            beta=beta,
            regularization=self.lambda_reg,
            alpha=0.0,
            is_regularized=is_regularized,
            beta_lower_bound=self.beta_lower_bound,
            beta_upper_bound=self.beta_upper_bound,
            selection=self.selection,
            tolerance=self.tolerance,
            max_iterations=self.max_iterations,
        )
        return beta

    def fit_beta_path(self, x_gram, y_gram, is_regularized):
        return super().fit_beta_path(x_gram, y_gram, is_regularized)

    def update_beta_path(self, x_gram, y_gram, beta_path, is_regularized):
        return super().update_beta_path(x_gram, y_gram, beta_path, is_regularized)



from itertools import product

import numpy as np
import pytest

import ondil

DISTRIBUTIONS = [
    getattr(ondil.distributions, name)() for name in ondil.distributions.__all__
]


@pytest.mark.parametrize(
    "distribution", DISTRIBUTIONS, ids=lambda dist: dist.__class__.__name__
)
def test_initial_values(distribution):
    n_params = distribution.n_params
    lower = np.clip(distribution.distribution_support[0], -1e3, 1e3)
    upper = np.clip(distribution.distribution_support[1], -1e3, 1e3)
    y = np.random.uniform(low=lower, high=upper, size=1000)
    theta = distribution.initial_values(y)
    assert theta.shape == (y.shape[0], n_params), "Initial values shape mismatch"
    assert np.all(
        distribution.parameter_support[param][0]
        <= theta[:, param]
        <= distribution.parameter_support[param][1]
        for param in range(n_params)
    ), "Initial values out of parameter support"


@pytest.mark.parametrize(
    "distribution", DISTRIBUTIONS, ids=lambda dist: dist.__class__.__name__
)
def test_raise_error_cross_derivative(distribution):
    n_params = distribution.n_params
    # We just want some values here on a reasonable space to ensure that
    # We can test the raise of the derivative
    lower = np.clip(distribution.distribution_support[0], -1e3, 1e3)
    upper = np.clip(distribution.distribution_support[1], -1e3, 1e3)
    y = np.random.uniform(low=lower, high=upper, size=1000)
    theta = distribution.initial_values(y)

    assert theta.shape == (y.shape[0], n_params)

    for a, b in product(range(n_params), range(n_params)):
        if a == b:
            with pytest.raises(
                ValueError, match="Cross derivatives must use different parameters."
            ):
                distribution.dl2_dpp(y, theta, (a, b))
        else:
            deriv = distribution.dl2_dpp(y, theta, (a, b))
            assert y.shape == deriv.shape, "Derivative shape should match y.shape"



import copy

from ..base import EstimationMethod
from .elasticnet import ElasticNetPath
from .lasso_path import LassoPath
from .recursive_least_squares import OrdinaryLeastSquares


class EstimationMethodFactory:
    def __init__(self):
        pass

    @staticmethod
    def from_string(method: str):
        if method == "lasso":
            return LassoPath()
        elif method == "ols":
            return OrdinaryLeastSquares()
        elif method == "elasticnet":
            return ElasticNetPath(alpha=0.5)
        else:
            raise ValueError(
                "Did not recognize method. Please provide ['ols', 'lasso', 'elasticnet']."
            )


def get_estimation_method(method: EstimationMethod | str):
    if isinstance(method, str):
        out = EstimationMethodFactory().from_string(method=method)
    elif isinstance(method, EstimationMethod):
        # This is a safeguard. The user might initialise only one concrete class
        # and pass it to all estimation methods in multi-method estimators.
        # We want each parameter to have its individual estimation method.
        out = copy.copy(method)
    else:
        raise ValueError("Method not recognized")
    return out



from typing import Tuple

import numpy as np
import scipy.stats as st

from ..base import Distribution, LinkFunction, ScipyMixin
from ..links import Identity, Log


class Logistic(ScipyMixin, Distribution):
    """
    The Logistic distribution with location and scale parameterization.

    The probability density function is:
    $$
    f(y | \\mu, \\sigma) = \\frac{\\exp\\left(-\\frac{y - \\mu}{\\sigma}\\right)}{\\sigma \\left(1 + \\exp\\left(-\\frac{y - \\mu}{\\sigma}\\right)\\right)^2}
    $$

    This distribution corresponds to the LO() distribution in GAMLSS.
    """

    corresponding_gamlss: str = "LO"
    parameter_names = {0: "mu", 1: "sigma"}
    parameter_support = {0: (-np.inf, np.inf), 1: (np.nextafter(0, 1), np.inf)}
    distribution_support = (-np.inf, np.inf)
    scipy_dist = st.logistic
    scipy_names = {"mu": "loc", "sigma": "scale"}

    def __init__(
        self,
        loc_link: LinkFunction = Identity(),
        scale_link: LinkFunction = Log(),
    ) -> None:
        super().__init__(links={0: loc_link, 1: scale_link})

    def theta_to_scipy_params(self, theta: np.ndarray) -> dict:
        mu = theta[:, 0]
        sigma = theta[:, 1]
        return {"loc": mu, "scale": sigma}

    def dl1_dp1(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma = self.theta_to_params(theta)
        z = (y - mu) / sigma
        expz = np.exp(z)

        if param == 0:
            return (1 / sigma) * (expz - 1) / (1 + expz)

        if param == 1:
            return (
                -1 / sigma
                - (y - mu) / sigma**2
                + 2 * ((y - mu) / sigma**2) * expz / (1 + expz)
            )

    def dl2_dp2(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        _, sigma = self.theta_to_params(theta)

        if param == 0:
            return -1 / (3 * sigma**2)

        if param == 1:
            return -(1 / (3 * sigma**2)) * (1 + (np.pi**2) / 3)

    def dl2_dpp(
        self, y: np.ndarray, theta: np.ndarray, params: Tuple[int, int] = (0, 1)
    ) -> np.ndarray:
        self._validate_dl2_dpp_inputs(y, theta, params)
        return np.zeros_like(y)

    def initial_values(self, y: np.ndarray) -> np.ndarray:
        initial_params = [np.mean(y), (np.sqrt(3) * np.std(y, ddof=1)) / np.sqrt(np.pi)]
        return np.tile(initial_params, (y.shape[0], 1))



from .factory import get_estimation_method
from .lasso_path import LassoPath
from .ridge import Ridge
from .elasticnet import ElasticNetPath
from .recursive_least_squares import OrdinaryLeastSquares

__all__ = [
    "get_estimation_method",
    "LassoPath",
    "Ridge",
    "ElasticNetPath",
    "OrdinaryLeastSquares",
]



import numpy as np

from ..base import LinkFunction


class Identity(LinkFunction):
    """
    The identity link function.

    The identity link is defined as \(g(x) = x\).
    """

    link_support = (-np.inf, np.inf)

    def __init__(self):
        pass

    def link(self, x: np.ndarray) -> np.ndarray:
        return x

    def inverse(self, x: np.ndarray) -> np.ndarray:
        return x

    def inverse_derivative(self, x: np.ndarray) -> np.ndarray:
        return np.ones_like(x)

    def link_derivative(self, x: np.ndarray) -> np.ndarray:
        return np.ones_like(x)

    def link_second_derivative(self, x: np.ndarray) -> np.ndarray:
        return np.zeros_like(x)



class OutOfSupportError(ValueError):
    """Exception raised for values that are out of support."""

    def __init__(self, message="This value is out of support."):
        self.message = message
        super().__init__(self.message)



import numpy as np

from ..base import LinkFunction
from .robust_math import robust_exp, robust_log


class Logit(LinkFunction):
    r"""The Logit Link function.

    The logit-link function is defined as \(g(x) = \log (x/ (1-x))\).
    """

    link_support = (np.nextafter(0, 1), np.nextafter(1, 0))

    def __init__(self):
        pass

    def link(self, x: np.ndarray) -> np.ndarray:
        return robust_log(x / (1 - x))

    def inverse(self, x: np.ndarray) -> np.ndarray:
        return robust_exp(x) / (robust_exp(x) + 1)

    def inverse_derivative(self, x: np.ndarray) -> np.ndarray:
        return robust_exp(x) / ((robust_exp(x) + 1) ** 2)

    def link_derivative(self, x: np.ndarray) -> np.ndarray:
        return 1 / (x * (1 - x))

    def link_second_derivative(self, x: np.ndarray) -> np.ndarray:
        return -(1 - 2 * x) / ((x**2) * ((1 - x) ** 2))



import numpy as np
import scipy.stats as st

from ..base import Distribution, LinkFunction, ScipyMixin
from ..links import Identity, Log


class Gumbel(ScipyMixin, Distribution):
    """
    The Gumbel distribution.

    The probability density function is given by:
    $$
        f(y|\\mu, \\sigma) = (1/\\sigma) * \\exp(-(z + \\exp(-z)))
    $$
    where $z = (y - \\mu)/\\sigma$ and has the following parameters:

    * $\\mu$: location
    * $\\sigma$: scale (>0)
    """

    corresponding_gamlss: str = "GU"
    parameter_names = {0: "mu", 1: "sigma"}
    parameter_support = {
        0: (-np.inf, np.inf),
        1: (np.nextafter(0, 1), np.inf),
    }
    distribution_support = (-np.inf, np.inf)
    scipy_dist = st.gumbel_l
    scipy_names = {"mu": "loc", "sigma": "scale"}

    def __init__(
        self,
        loc_link: LinkFunction = Identity(),
        scale_link: LinkFunction = Log(),
    ) -> None:
        super().__init__(links={0: loc_link, 1: scale_link})

    def dl1_dp1(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        mu, sigma = self.theta_to_params(theta)
        z = (y - mu) / sigma
        if param == 0:
            # d logL / d mu
            # (exp((y-mu)/sigma)-1)/sigma,
            return (np.exp(z) - 1) / sigma
        if param == 1:
            # d logL / d sigma#
            # -(1/sigma)+  ((y-mu)/sigma2)*(exp((y-mu)/sigma)-1),
            return -(1 / sigma) + (z / sigma) * (np.exp(z) - 1)
        raise ValueError("param must be 0 (mu) or 1 (sigma)")

    def dl2_dp2(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        _, sigma = self.theta_to_params(theta)
        if param == 0:
            # d^2 logL / d mu^2
            return -1 / sigma**2
        if param == 1:
            # d^2 logL / d sigma^2
            return -1.82368 / sigma**2
        raise ValueError("param must be 0 (mu) or 1 (sigma)")

    def dl2_dpp(self, y: np.ndarray, theta: np.ndarray, params=(0, 1)) -> np.ndarray:
        self._validate_dl2_dpp_inputs(y, theta, params)
        _, sigma = self.theta_to_params(theta)
        if sorted(params) == [0, 1]:
            # d^2 logL / d mu d sigma
            # Note: order does not matter for cross-derivative
            return -0.422784 / sigma**2
        raise ValueError("Cross derivatives must use different parameters.")

    def initial_values(self, y: np.ndarray) -> np.ndarray:
        out = np.zeros((y.shape[0], 2))
        out[:, 0] = np.mean(y, axis=0)
        out[:, 1] = np.sqrt(6) * np.std(y, axis=0) / np.pi
        return out



import numbers
import warnings
from typing import Literal, Optional

import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin, _fit_context
from sklearn.utils._param_validation import Interval, StrOptions
from sklearn.utils.multiclass import type_of_target
from sklearn.utils.validation import (
    _check_sample_weight,
    check_is_fitted,
    validate_data,
)

from ..base import EstimationMethod, OndilEstimatorMixin
from ..design_matrix import add_intercept
from ..information_criteria import InformationCriterion
from ..methods import get_estimation_method
from ..scaler import OnlineScaler
from ..utils import calculate_effective_training_length


class OnlineLinearModel(OndilEstimatorMixin, RegressorMixin, BaseEstimator):
    "Simple Online Linear Regression for the expected value."

    _parameter_constraints = {
        "forget": [Interval(numbers.Real, 0.0, 1.0, closed="left")],
        "fit_intercept": [bool],
        "scale_inputs": [bool, np.ndarray],
        "regularize_intercept": [bool],
        "method": [EstimationMethod, str],
        "ic": [StrOptions({"aic", "bic", "hqc", "max"})],
    }

    def __init__(
        self,
        forget: float = 0.0,
        scale_inputs: bool | np.ndarray = True,
        fit_intercept: bool = True,
        regularize_intercept: bool = False,
        method: EstimationMethod | str = "ols",
        ic: Literal["aic", "bic", "hqc", "max"] = "bic",
    ):
        """The basic linear model for many different estimation techniques.

        Args:
            forget (float, optional): Exponential discounting of old observations. Defaults to 0.
            scale_inputs (bool, optional): Whether to scale the $X$ matrix. Defaults to True.
            fit_intercept (bool, optional): Whether to add an intercept in the estimation. Defaults to True.
            regularize_intercept (bool, optional): Whether to regularize the intercept. Defaults to False.
            method (EstimationMethod | str, optional): The estimation method. Can be a string or `EstimationMethod` class. Defaults to "ols".
            ic (Literal["aic", "bic", "hqc", "max"], optional): The information criteria for model selection. Defaults to "bic".
        Raises:
            ValueError: Will raise if you try to regularize the intercept, but not fit it.
        """

        self.forget = forget
        self.method = method
        self.fit_intercept = fit_intercept
        self.scale_inputs = scale_inputs
        self.regularize_intercept = regularize_intercept
        self.ic = ic

    @property
    def beta(self):
        check_is_fitted(self)
        return self.coef_

    @property
    def beta_path(self):
        check_is_fitted(self)
        return self.coef_path_

    def __sklearn_tags__(self):
        tags = super().__sklearn_tags__()
        return tags

    def _prepare_fit(self):
        self._method = get_estimation_method(self.method)
        self._scaler = OnlineScaler(forget=self.forget, to_scale=self.scale_inputs)

        # Raise on inputs that does not make sense
        if not self._method._path_based_method and self.regularize_intercept:
            warnings.warn(
                "Note that you have passed a non-regularized estimation method but want to regularize the intercept. Are you sure?"
            )
        if self.regularize_intercept and not self.fit_intercept:
            raise ValueError(
                "You want to regularize the intercept but not fit an intercept."
            )

    def get_design_matrix(self, X: np.ndarray):
        if self.fit_intercept:
            design = add_intercept(X)
        else:
            design = np.copy(X)
        return design

    @_fit_context(prefer_skip_nested_validation=True)
    def fit(
        self,
        X: np.ndarray,
        y: np.ndarray,
        sample_weight: Optional[np.ndarray] = None,
    ) -> "OnlineLinearModel":
        """Initial fit of the online regression model.

        Args:
            X (np.ndarray): The design matrix $X$.
            y (np.ndarray): The response vector $y$.
            sample_weight (Optional[np.ndarray], optional): The sample weights. Defaults to None.
        """

        self._prepare_fit()
        X, y = validate_data(self, X=X, y=y, reset=True, dtype=[np.float64, np.float32])
        _ = type_of_target(y, raise_unknown=True)
        sample_weight = _check_sample_weight(X=X, sample_weight=sample_weight)

        self.n_observations_ = np.sum(sample_weight)
        self.n_features_ = X.shape[1] + int(self.fit_intercept)

        if self.n_observations_ <= self.n_features_ and self.method == "ols":
            raise ValueError(
                f"You have tried to fit using n_samples={self.n_observations_} and n_features={self.n_features_}. "
                "Since we need calculate the inverse Gram matrix, we need at least as many observations as features for OLS-based methods"
            )
        if self.n_observations_ == 1:
            raise ValueError(
                "You have tried to fit using only one sample. This is not supported for online linear regression."
            )

        self.n_training_ = calculate_effective_training_length(
            self.forget, self.n_observations_
        )

        self._scaler.fit(X, sample_weight=sample_weight)
        X_scaled = self._scaler.transform(X)
        X_scaled = self.get_design_matrix(X=X_scaled)

        self._is_regularized = np.repeat(True, self.n_features_)
        if self.fit_intercept:
            self._is_regularized[0] = self.regularize_intercept

        self._x_gram = self._method.init_x_gram(
            X_scaled, weights=sample_weight, forget=self.forget
        )
        self._y_gram = self._method.init_y_gram(
            X_scaled, y, weights=sample_weight, forget=self.forget
        )

        if self._method._path_based_method:
            self.coef_path_ = self._method.fit_beta_path(
                x_gram=self._x_gram,
                y_gram=self._y_gram,
                is_regularized=self._is_regularized,
            )
            residuals = np.expand_dims(sample_weight, -1) * (
                np.expand_dims(y, -1) - X_scaled @ self.coef_path_.T
            )
            self._rss = np.sum(residuals**2, axis=0)
            n_params = np.sum(~np.isclose(self.coef_path_, 0), axis=1)
            ic = InformationCriterion(
                n_observations=self.n_training_,
                n_parameters=n_params,
                criterion=self.ic,
            ).from_rss(rss=self._rss)
            best_ic = np.argmin(ic)
            self.coef_ = self.coef_path_[best_ic, :]
        else:
            self.coef_ = self._method.fit_beta(
                self._x_gram, self._y_gram, self._is_regularized
            )

        return self

    @_fit_context(prefer_skip_nested_validation=True)
    def update(
        self,
        X: np.ndarray,
        y: np.ndarray,
        sample_weight: Optional[np.ndarray] = None,
    ) -> None:
        """Update the regression model.

        Args:
            X (np.ndarray): The new row of the design matrix $X$. Needs to be of shape 1 x n_features or n_obs_new x n_features.
            y (np.ndarray): The new observation of $y$. Needs to be the same shape as `X` or a single observation.
            sample_weight (Optional[np.ndarray], optional): The weight for the new observations. `None` implies all observations have weight 1. Defaults to None.
        """

        X, y = validate_data(
            self, X=X, y=y, reset=False, dtype=[np.float64, np.float32]
        )
        sample_weight = _check_sample_weight(X=X, sample_weight=sample_weight)
        _ = type_of_target(y, raise_unknown=True)

        self.n_observations_ += sample_weight.sum()
        self.n_training_ = calculate_effective_training_length(
            self.forget, self.n_observations_
        )

        self._scaler.update(X, sample_weight=sample_weight)
        X_scaled = self._scaler.transform(X)
        X_scaled = self.get_design_matrix(X=X_scaled)

        self._x_gram = self._method.update_x_gram(
            self._x_gram, X_scaled, forget=self.forget, weights=sample_weight
        )
        self._y_gram = self._method.update_y_gram(
            self._y_gram, X_scaled, y, forget=self.forget, weights=sample_weight
        )
        if self._method._path_based_method:
            self.coef_path_ = self._method.update_beta_path(
                x_gram=self._x_gram,
                y_gram=self._y_gram,
                beta_path=self.coef_path_,
                is_regularized=self._is_regularized,
            )

            residuals = np.expand_dims(y, -1) - X_scaled @ self.coef_path_.T
            self._rss = (1 - self.forget) * self._rss + np.sum(residuals**2, axis=0)
            n_params = np.sum(~np.isclose(self.coef_path_, 0), axis=1)
            ic = InformationCriterion(
                n_observations=self.n_training_,
                n_parameters=n_params,
                criterion=self.ic,
            ).from_rss(rss=self._rss)
            best_ic = np.argmin(ic)
            self.coef_ = self.coef_path_[best_ic, :]
        else:
            self.coef_ = self._method.update_beta(
                x_gram=self._x_gram,
                y_gram=self._y_gram,
                beta=self.coef_,
                is_regularized=self._is_regularized,
            )

        return self

    def score(self, X: np.ndarray, y: np.ndarray) -> float:
        """Calculate the coefficient of determination $R^2$.

        Args:
            X (np.ndarray): The design matrix $X$.
            y (np.ndarray): The response vector $y$.

        Returns:
            float: The coefficient of determination $R^2$.
        """
        check_is_fitted(self)
        X, y = validate_data(
            self, X=X, y=y, reset=False, dtype=[np.float64, np.float32]
        )

        prediction = self.predict(X)
        ss_residuals = np.sum((y - prediction) ** 2)
        ss_total = np.sum((y - np.mean(y)) ** 2)

        return 1 - ss_residuals / ss_total

    def predict(self, X: np.ndarray) -> np.ndarray:
        """Predict using the optimal IC selection.

        Args:
            X (np.ndarray): The design matrix $X$.

        Returns:
            np.ndarray: The predictions for the optimal IC.
        """
        check_is_fitted(self)
        X = validate_data(self, X=X, reset=False, dtype=[np.float64, np.float32])

        design = self._scaler.transform(X)
        design = self.get_design_matrix(X=design)
        prediction = design @ self.coef_.T
        return prediction

    def predict_path(self, X: np.ndarray) -> np.ndarray:
        """Predict the full regularization path.

        Args:
            X (np.ndarray): The design matrix $X$.

        Returns:
            np.ndarray: The predictions for the full path.
        """
        check_is_fitted(self)
        X = validate_data(self, X=X, reset=False, dtype=[np.float64, np.float32])

        design = self._scaler.transform(X)
        design = self.get_design_matrix(X=design)
        prediction = design @ self.coef_path_.T
        return prediction



from abc import ABC, abstractmethod
from typing import Tuple

import numpy as np


class LinkFunction(ABC):
    """The base class for the link functions."""

    @property
    @abstractmethod
    def link_support(self) -> Tuple[float, float]:
        """The support of the distribution."""
        pass

    @abstractmethod
    def link(self, x: np.ndarray) -> np.ndarray:
        """Calculate the Link"""

    @abstractmethod
    def inverse(self, x: np.ndarray) -> np.ndarray:
        """Calculate the inverse of the link function"""

    @abstractmethod
    def link_derivative(self, x: np.ndarray) -> np.ndarray:
        """Calculate the first derivative of the link function"""
        raise NotImplementedError("Currently not implemented. Will be needed for GLMs")

    @abstractmethod
    def link_second_derivative(self, x: np.ndarray) -> np.ndarray:
        """Calculate the second derivative for the link function"""
        raise NotImplementedError("Currently not implemented.")

    @abstractmethod
    def inverse_derivative(self, x: np.ndarray) -> np.ndarray:
        """Calculate the first derivative for the inverse link function"""
        raise NotImplementedError("Currently not implemented.")
